import pandas as pd
import numpy as np
import Bio
from Bio import SeqIO
from Bio.Seq import Seq
from pandarallel import pandarallel
import matplotlib.pyplot as plt
from collections import Counter, defaultdict
from typeguard import typechecked
from datetime import date
import re
from multiprocessing import Pool
from functools import partial
from itertools import repeat
import gzip
import random
from enum import Enum
from typing import Callable
from typing import Union, List, Mapping, Tuple, Optional, Any, DefaultDict
from typing import Counter as CounterType
from concurrent.futures import ProcessPoolExecutor

from . import crispr_sequence_encoding
from ..models.mapping_models import CompleteInferenceMatchResult, MatchSetSingleInferenceMatchResult, MatchSetSingleInferenceMatchResultValue, SurrogateProtospacerMismatchSingleInferenceMatchResult, SurrogateProtospacerMismatchSingleInferenceMatchResultValue
from ..models.error_models import (GuideCountErrorType, 
                                   ProtospacerHammingThresholdGuideCountError, SurrogateHammingThresholdGuideCountError, BarcodeHammingThresholdGuideCountError,
                                   MissingInfoGuideCountError, InsufficientLengthGuideCountError, 
                                   ProtospacerMissingInfoGuideCountError, SurrogateMissingInfoGuideCountError, BarcodeMissingInfoGuideCountError, 
                                   ProtospacerInsufficientLengthGuideCountError, SurrogateInsufficientLengthGuideCountError, BarcodeInsufficientLengthGuideCountError)


@typechecked
def infer_whitelist_sequence(observed_guide_reporter_sequence_input: Tuple[str, Optional[str], Optional[str]], 
        whitelist_guide_reporter_df: pd.DataFrame, 
        contains_surrogate:bool, 
        contains_barcode:bool, 
        contains_umi:bool, 
        encoded_whitelist_protospacer_sequences_series: np.ndarray, 
        encoded_whitelist_surrogate_sequences_series: Optional[np.ndarray] = None, 
        encoded_whitelist_barcode_sequences_series: Optional[np.ndarray] = None, 
        protospacer_hamming_threshold: int = 7, 
        surrogate_hamming_threshold: int = 10, 
        barcode_hamming_threshold: int = 2):

    #
    #   FORMAT OBSERVED SEQUENCES AS PANDAS SERIES
    #
    observed_reporter_sequences_indices = ["protospacer"]
    if contains_surrogate:
        observed_reporter_sequences_indices.append("surrogate")
    if contains_barcode:
        observed_reporter_sequences_indices.append("barcode")
    observed_guide_reporter_sequence = pd.Series(observed_guide_reporter_sequence_input, index=observed_reporter_sequences_indices)
    

    #
    # QUICK VALIDATION OF THE OBSERVED SEQUENCES, SAVE RESULTS AS VARIABLES TO BE USED DOWNSTREAM
    #
    def validate_observed_sequence(sequence_name: str, encoded_whitelist_sequence_series: np.ndarray, missing_info_error: MissingInfoGuideCountError, insufficient_length_error: InsufficientLengthGuideCountError):
        if (observed_guide_reporter_sequence[sequence_name] is None) or (observed_guide_reporter_sequence[sequence_name] == "None") or (observed_guide_reporter_sequence[sequence_name].strip() == ""):
            return missing_info_error
        
        if len(observed_guide_reporter_sequence[sequence_name]) < insufficient_length_error.minimum_length: 
            return insufficient_length_error   
        
        return None
        
    protospacer_error_result = validate_observed_sequence(sequence_name="protospacer", encoded_whitelist_sequence_series=encoded_whitelist_protospacer_sequences_series, missing_info_error=ProtospacerMissingInfoGuideCountError(sequence_value=observed_guide_reporter_sequence), insufficient_length_error=ProtospacerInsufficientLengthGuideCountError(sequence_length=len(observed_guide_reporter_sequence["protospacer"]), minimum_length=whitelist_guide_reporter_df["protospacer"].apply(len).min()))
    surrogate_error_result = None
    if contains_surrogate:
        surrogate_error_result = validate_observed_sequence(sequence_name="surrogate", encoded_whitelist_sequence_series=encoded_whitelist_surrogate_sequences_series, missing_info_error=SurrogateMissingInfoGuideCountError(sequence_value=observed_guide_reporter_sequence), insufficient_length_error=SurrogateInsufficientLengthGuideCountError(sequence_length=len(observed_guide_reporter_sequence["surrogate"]), minimum_length=whitelist_guide_reporter_df["surrogate"].apply(len).min()))
    barcode_error_result = None
    if contains_barcode:
        barcode_error_result = validate_observed_sequence(sequence_name="barcode", encoded_whitelist_sequence_series=encoded_whitelist_barcode_sequences_series, missing_info_error=BarcodeMissingInfoGuideCountError(sequence_value=observed_guide_reporter_sequence), insufficient_length_error=BarcodeInsufficientLengthGuideCountError(sequence_length=len(observed_guide_reporter_sequence["barcode"]), minimum_length=whitelist_guide_reporter_df["barcode"].apply(len).min()))

    # CHANGE SHAPE OF ENCODINGS TO BE SAME LENGTH
    if protospacer_error_result is None:
        observed_guide_reporter_sequence["protospacer"] = observed_guide_reporter_sequence["protospacer"][:encoded_whitelist_protospacer_sequences_series.shape[1]]
        encoded_whitelist_protospacer_sequences_series = encoded_whitelist_protospacer_sequences_series[:, :len(observed_guide_reporter_sequence["protospacer"]), :] # Change shape of whitelist library encodings
    if contains_surrogate and (surrogate_error_result is None):
        #print(len(observed_guide_reporter_sequence["surrogate"]))
        #print(len(encoded_whitelist_surrogate_sequences_series.shape))
        observed_guide_reporter_sequence["surrogate"] = observed_guide_reporter_sequence["surrogate"][:encoded_whitelist_protospacer_sequences_series.shape[1]]
        encoded_whitelist_surrogate_sequences_series = encoded_whitelist_surrogate_sequences_series[:, :len(observed_guide_reporter_sequence["surrogate"]), :] # Change shape of whitelist library encodings
        #print(len(observed_guide_reporter_sequence["surrogate"]))
        #print(len(encoded_whitelist_surrogate_sequences_series.shape))
        #print("Done")
    if contains_barcode and (barcode_error_result is None):
        observed_guide_reporter_sequence["barcode"] = observed_guide_reporter_sequence["barcode"][:encoded_whitelist_protospacer_sequences_series.shape[1]]
        encoded_whitelist_barcode_sequences_series = encoded_whitelist_barcode_sequences_series[:, :len(observed_guide_reporter_sequence["barcode"]), :] # Change shape of whitelist library encodings
    
    # PERFORM MAPPPING ON BOTH THE PROTOSPACER-ONLY, FULL, AND SURROGATE MISMAPPED-PROTOSPACER.
    complete_match_result = CompleteInferenceMatchResult()

    #
    # GET THE BARCODE-ONLY MATCHES
    # (Since barcode-matching will be done a lot, let's do first and save the match results)
    if contains_barcode: # IF BARCODE IS PARSED
        barcode_available = True # Let's save an indicator on if we successfully parsed barcode for convenience downstream
        if barcode_error_result is None: # IF NO BARCODE PARSING ISSUE

            #
            # FIND BARCODE MATCHES
            #
            try:
                observed_barcode_sequence_encoded = crispr_sequence_encoding.encode_DNA_base_vectorized(crispr_sequence_encoding.numpify_string_vectorized(observed_guide_reporter_sequence["barcode"]))  # Encode the observed barcode
            except Exception as e:
                print(e)
                raise Exception(f"Error encoding barcode sequence: {observed_guide_reporter_sequence['barcode']}")
            observed_barcode_sequence_dists = crispr_sequence_encoding.retrieve_hamming_distance_whitelist(observed_barcode_sequence_encoded, encoded_whitelist_barcode_sequences_series) # Retrieve hamming distance with whitelist barcode
            observed_barcode_sequence_dists_min = observed_barcode_sequence_dists.min() # Get the barcode with the minimum  hamming distance
            barcode_hamming_threshold_met = (observed_barcode_sequence_dists_min < barcode_hamming_threshold)

            barcode_matches_indices = np.where(observed_barcode_sequence_dists == observed_barcode_sequence_dists_min)[0] # Get the indices of ALL barcode matches
            whitelist_guide_reporter_df_barcode_match = whitelist_guide_reporter_df.iloc[barcode_matches_indices] # Get the whitelist reporters with the matched barcode(s)
            
            if barcode_hamming_threshold_met: # IF BARCODE MATCH
                encoded_whitelist_protospacer_sequences_series_barcode_match = encoded_whitelist_protospacer_sequences_series[barcode_matches_indices] # Subset the protospacer encodings with the barcode matches for later
                if contains_surrogate:
                    encoded_whitelist_surrogate_sequences_series_barcode_match = encoded_whitelist_surrogate_sequences_series[barcode_matches_indices] # Subset the surrogate encodings with the barcode matches for later
            else: # NO BARCODE MATCH, ERROR
                barcode_available = False
                barcode_error_result = BarcodeHammingThresholdGuideCountError(
                    hamming_min=observed_barcode_sequence_dists_min, 
                    hamming_threshold=barcode_hamming_threshold,
                    original_df=whitelist_guide_reporter_df,
                    hamming_min_match_df=whitelist_guide_reporter_df_barcode_match)
        else:
            barcode_available = False
            

    #
    # PREPARE THE PROTOSPACER-ONLYMATCHES
    #
    if protospacer_error_result is None:
        try:
            observed_protospacer_sequence_encoded = crispr_sequence_encoding.encode_DNA_base_vectorized(crispr_sequence_encoding.numpify_string_vectorized(observed_guide_reporter_sequence["protospacer"]))  # Encode the observed protospacer
        except Exception as e:
            print(e)
            raise Exception(f"Error encoding protospacer sequence: {observed_guide_reporter_sequence['protospacer']}")
        observed_protospacer_sequence_dists = crispr_sequence_encoding.retrieve_hamming_distance_whitelist(observed_protospacer_sequence_encoded, encoded_whitelist_protospacer_sequences_series) # Hamming distance among all whitelist protospacers
        observed_protospacer_sequence_dists_min = observed_protospacer_sequence_dists.min() # Get minimum hamming distance
        protospacer_hamming_threshold_met = (observed_protospacer_sequence_dists_min < protospacer_hamming_threshold)

        protospacer_matches_indices = np.where(observed_protospacer_sequence_dists == observed_protospacer_sequence_dists_min)[0]
        whitelist_guide_reporter_df_hamming_protospacer_match = whitelist_guide_reporter_df.iloc[protospacer_matches_indices] # Get all whitelisted guides with the minimum hamming distance (could be multiple)
        if protospacer_hamming_threshold_met: # PROTOSPACER MATCHED
            #
            # SET THE PROTOSPACER-ONLYMATCHES
            #
            encoded_whitelist_protospacer_sequences_series_protospacer_match = encoded_whitelist_protospacer_sequences_series[protospacer_matches_indices] # Subset the protospacer encodings with the protospacer matches for later
            complete_match_result.protospacer_match = MatchSetSingleInferenceMatchResult(value=MatchSetSingleInferenceMatchResultValue(matches=whitelist_guide_reporter_df_hamming_protospacer_match))
            
            if contains_barcode: # IF BARCODE IS PARSED, PROCEED TO BARCODE-MATCHING INFERENCE
                if barcode_available: # IF BARCODE IS MATCHED, PROCEED TO BARCODE-MATCHING INFERENCE
                    #
                    # PREPARE PROTOSPACER-MATCH, BARCODE-MATCH
                    #
                    observed_protospacer_sequence_dists_barcode_match = crispr_sequence_encoding.retrieve_hamming_distance_whitelist(observed_protospacer_sequence_encoded, encoded_whitelist_protospacer_sequences_series_barcode_match) # Hamming distance among barcode-match protospacers
                    observed_protospacer_sequence_dists_barcode_match_min = observed_protospacer_sequence_dists_barcode_match.min() # Get minimum hamming distance
                    barcode_match_protospacer_hamming_threshold_met = (observed_protospacer_sequence_dists_barcode_match_min < protospacer_hamming_threshold)

                    barcode_match_protospacer_match_indices = np.where(observed_protospacer_sequence_dists_barcode_match == observed_protospacer_sequence_dists_barcode_match_min)[0]
                    whitelist_guide_reporter_df_hamming_barcode_match_protospacer_match = whitelist_guide_reporter_df_barcode_match.iloc[barcode_match_protospacer_match_indices] # Get all whitelisted guides with the minimum hamming distance (could be multiple)
                    if barcode_match_protospacer_hamming_threshold_met:  # PROTOSPACER MATCHED WHEN BARCODE IS SPECIFIED
                        #
                        # SET PROTOSPACER-MATCH, BARCODE-MATCH
                        #
                        
                        complete_match_result.protospacer_match_barcode_match = MatchSetSingleInferenceMatchResult(value=MatchSetSingleInferenceMatchResultValue(matches=whitelist_guide_reporter_df_hamming_barcode_match_protospacer_match))
                        if contains_surrogate: # IF SURROGATE IS PARSED, PROCEED TO SURROGATE-MATCHING INFERENCE
                            if surrogate_error_result is None:
                                #
                                # PREPARE PROTOSPACER-MATCH, BARCODE-MATCH, SURROGATE-MATCH
                                #
                                encoded_whitelist_surrogate_sequences_series_barcode_match_protospacer_match = encoded_whitelist_surrogate_sequences_series_barcode_match[barcode_match_protospacer_match_indices] # Subset the surrogate encodings with the protospacer and encoding matches for later
                                try:
                                    observed_surrogate_sequence_encoded = crispr_sequence_encoding.encode_DNA_base_vectorized(crispr_sequence_encoding.numpify_string_vectorized(observed_guide_reporter_sequence["surrogate"]))  # Encode the observed protospacer
                                except Exception as e:
                                    print(f"Error message: {e}")
                                    raise Exception(f"Error encoding surrogate sequence: {observed_guide_reporter_sequence['surrogate']}")
                                observed_surrogate_sequence_dists_barcode_match_protospacer_match = crispr_sequence_encoding.retrieve_hamming_distance_whitelist(observed_surrogate_sequence_encoded, encoded_whitelist_surrogate_sequences_series_barcode_match_protospacer_match) # Hamming distance among all whitelist sub-selected surrogates
                                observed_surrogate_sequence_dists_barcode_match_protospacer_match_min = observed_surrogate_sequence_dists_barcode_match_protospacer_match.min()
                                barcode_match_protospacer_match_surrogate_hamming_threshold_met = (observed_surrogate_sequence_dists_barcode_match_protospacer_match_min < surrogate_hamming_threshold)
                                
                                barcode_match_protospacer_match_surrogate_match_indices = np.where(observed_surrogate_sequence_dists_barcode_match_protospacer_match == observed_surrogate_sequence_dists_barcode_match_protospacer_match_min)[0]
                                whitelist_guide_reporter_df_hamming_barcode_match_protospacer_match_surrogate_match = whitelist_guide_reporter_df_hamming_barcode_match_protospacer_match.iloc[barcode_match_protospacer_match_surrogate_match_indices] # Get all whitelisted guides with the minimum hamming distance (could be multiple)
                                if barcode_match_protospacer_match_surrogate_hamming_threshold_met: # SURROGATE MATCHED
                                    #
                                    # SET PROTOSPACER-MATCH, BARCODE-MATCH, SURROGATE-MATCH
                                    #
                                    complete_match_result.protospacer_match_surrogate_match_barcode_match = MatchSetSingleInferenceMatchResult(value=MatchSetSingleInferenceMatchResultValue(matches=whitelist_guide_reporter_df_hamming_barcode_match_protospacer_match_surrogate_match))
                                else: # NO SURROGATE MATCH, ERROR
                                    complete_match_result.protospacer_match_surrogate_match_barcode_match = MatchSetSingleInferenceMatchResult(
                                        error=SurrogateHammingThresholdGuideCountError(
                                            hamming_min=observed_surrogate_sequence_dists_barcode_match_protospacer_match_min, 
                                            hamming_threshold=surrogate_hamming_threshold,
                                            original_df=whitelist_guide_reporter_df_hamming_barcode_match_protospacer_match,
                                            hamming_min_match_df=whitelist_guide_reporter_df_hamming_barcode_match_protospacer_match_surrogate_match,
                                            barcode_subsetted=True))
                            else: # PARSED SURROGATE HAS SOME UNEXPECTED ISSUE 
                                complete_match_result.protospacer_match_surrogate_match_barcode_match = MatchSetSingleInferenceMatchResult(error=surrogate_error_result)
                    else: # NO PROTOSPACER MATCH, ERROR
                        complete_match_result.protospacer_match_barcode_match = MatchSetSingleInferenceMatchResult(
                                        error=ProtospacerHammingThresholdGuideCountError(
                                            hamming_min=observed_protospacer_sequence_dists_barcode_match_min, 
                                            hamming_threshold=protospacer_hamming_threshold,
                                            original_df=whitelist_guide_reporter_df_barcode_match,
                                            hamming_min_match_df=whitelist_guide_reporter_df_hamming_barcode_match_protospacer_match,
                                            barcode_subsetted=True))
                else: # NO BARCODE MATCH, SET ERROR TO THE RESULTS REQUIREING A BARCODE MATCH
                    complete_match_result.protospacer_match_barcode_match = MatchSetSingleInferenceMatchResult(error=barcode_error_result)
                    complete_match_result.protospacer_match_surrogate_match_barcode_match = MatchSetSingleInferenceMatchResult(error=barcode_error_result)

            #
            # PROTOSPACER-MATCH, SURROGATE-MATCH, NO BARCODE
            #
            if contains_surrogate: # IF SURROGATE IS PARSED
                if surrogate_error_result is None: # AND NO SURROGATE PARSING ISSUE 

                    #
                    # GET SURROGATE-MATCHES ON THE PROTOSPACER-MATCHES
                    #
                    encoded_whitelist_surrogate_sequences_series_protospacer_match = encoded_whitelist_surrogate_sequences_series[protospacer_matches_indices]
                    try:
                        observed_surrogate_sequence_encoded = crispr_sequence_encoding.encode_DNA_base_vectorized(crispr_sequence_encoding.numpify_string_vectorized(observed_guide_reporter_sequence["surrogate"]))  # Encode the observed protospacer
                    except Exception as e:
                        print(e)
                        raise Exception(f"Error encoding surrogate sequence: {observed_guide_reporter_sequence['surrogate']}")
                    observed_surrogate_sequence_dists_protospacer_match = crispr_sequence_encoding.retrieve_hamming_distance_whitelist(observed_surrogate_sequence_encoded, encoded_whitelist_surrogate_sequences_series_protospacer_match) # TODO: This should be on the protospacer-match array
                    observed_surrogate_sequence_dists_protospacer_match_min = observed_surrogate_sequence_dists_protospacer_match.min()
                    protospacer_match_surrogate_hamming_threshold_met = (observed_surrogate_sequence_dists_protospacer_match_min < surrogate_hamming_threshold)

                    protospacer_match_surrogate_match_indices = np.where(observed_surrogate_sequence_dists_protospacer_match == observed_surrogate_sequence_dists_protospacer_match_min)[0]
                    whitelist_guide_reporter_df_hamming_protospacer_match_surrogate_match = whitelist_guide_reporter_df_hamming_protospacer_match.iloc[protospacer_match_surrogate_match_indices] # Get all whitelisted guides with the minimum hamming distance (could be multiple)
                    if protospacer_match_surrogate_hamming_threshold_met: # IF SURROGATE MATCH
                        complete_match_result.protospacer_match_surrogate_match = MatchSetSingleInferenceMatchResult(value=MatchSetSingleInferenceMatchResultValue(matches=whitelist_guide_reporter_df_hamming_protospacer_match_surrogate_match))
                    else: # NO SURROGATE MATCH, ERROR OUT
                        complete_match_result.protospacer_match_surrogate_match = MatchSetSingleInferenceMatchResult(
                                        error=SurrogateHammingThresholdGuideCountError(
                                            hamming_min=observed_surrogate_sequence_dists_protospacer_match_min, 
                                            hamming_threshold=surrogate_hamming_threshold,
                                            original_df=whitelist_guide_reporter_df_hamming_protospacer_match,
                                            hamming_min_match_df=whitelist_guide_reporter_df_hamming_protospacer_match_surrogate_match,
                                            barcode_subsetted=False))
                else: # SURROGATE PARSING ISSUE, ERROR
                    complete_match_result.protospacer_match_surrogate_match = MatchSetSingleInferenceMatchResult(error=surrogate_error_result)



            #
            # GET THE PROTOSPACER-SURROGATE MISMATCHES NOW (if surrogate is available)
            #
            if contains_surrogate: # IF SURROGATE IS PARSED
                if surrogate_error_result is None: # AND NO SURROGATE PARSING ISSUE
                    if contains_barcode: # IF BARCODE IS PARSED
                        if barcode_available: # AND NO BARCODE PARSING ISSUE
                            #
                            # PREPARE SURROGATE-MATCH, BARCODE-MATCH
                            #
                            try:
                                observed_surrogate_sequence_encoded = crispr_sequence_encoding.encode_DNA_base_vectorized(crispr_sequence_encoding.numpify_string_vectorized(observed_guide_reporter_sequence["surrogate"]))  # Encode the observed protospacer
                            except Exception as e:
                                print(e)
                                raise Exception(f"Error encoding surrogate sequence: {observed_guide_reporter_sequence['surrogate']}")
                            observed_surrogate_sequence_dists_barcode_match = crispr_sequence_encoding.retrieve_hamming_distance_whitelist(observed_surrogate_sequence_encoded, encoded_whitelist_surrogate_sequences_series_barcode_match) # Hamming distance among barcode-match protospacers
                            observed_surrogate_sequence_dists_barcode_match_min = observed_surrogate_sequence_dists_barcode_match.min() # Get minimum hamming distance
                            barcode_match_surrogate_hamming_threshold_met = (observed_surrogate_sequence_dists_barcode_match_min < surrogate_hamming_threshold)
                            
                             #
                            # SET SURROGATE-MATCH, BARCODE-MATCH
                            #
                            barcode_match_surrogate_match_indices = np.where(observed_surrogate_sequence_dists_barcode_match == observed_surrogate_sequence_dists_barcode_match_min)[0]
                            whitelist_guide_reporter_df_hamming_barcode_match_surrogate_match = whitelist_guide_reporter_df_barcode_match.iloc[barcode_match_surrogate_match_indices] # Get all whitelisted guides with the minimum hamming distance (could be multiple)

                            if barcode_match_surrogate_hamming_threshold_met: # If the surrogate matches are below the surrogate hamming threshold
                                # See if there are identical matches between protospacer-only matches and surrogate-only matches
                                whitelist_guide_reporter_df_hamming_barcode_match_surrogate_match_protospacer_match = pd.merge(whitelist_guide_reporter_df_hamming_barcode_match_surrogate_match, whitelist_guide_reporter_df_hamming_protospacer_match, how='inner')
                                
                                complete_match_result.protospacer_mismatch_surrogate_match_barcode_match = SurrogateProtospacerMismatchSingleInferenceMatchResult(value=SurrogateProtospacerMismatchSingleInferenceMatchResultValue(
                                    mismatched=whitelist_guide_reporter_df_hamming_barcode_match_surrogate_match_protospacer_match.empty, 
                                    surrogate_matches=whitelist_guide_reporter_df_hamming_barcode_match_surrogate_match, 
                                    protospacer_matches=whitelist_guide_reporter_df_hamming_protospacer_match,
                                    protospacer_surrogate_matches=whitelist_guide_reporter_df_hamming_barcode_match_surrogate_match_protospacer_match
                                ))
                            else: # NO SURROGATE MATCH
                                complete_match_result.protospacer_mismatch_surrogate_match_barcode_match = SurrogateProtospacerMismatchSingleInferenceMatchResult(
                                    error=SurrogateHammingThresholdGuideCountError(
                                        hamming_min=observed_surrogate_sequence_dists_barcode_match_min,
                                        hamming_threshold=surrogate_hamming_threshold,
                                        original_df=whitelist_guide_reporter_df_barcode_match,
                                        hamming_min_match_df=whitelist_guide_reporter_df_hamming_barcode_match_surrogate_match,
                                        barcode_subsetted=True,
                                    ))
                        else: # BARCODE PARSING ISSUE OR NO MATCH, ERROR OUT
                            complete_match_result.protospacer_mismatch_surrogate_match_barcode_match = SurrogateProtospacerMismatchSingleInferenceMatchResult(error=barcode_error_result)
                    #
                    # PREPARE SURROGATE MATCHES ON THE PROTOSPACER-MATCHES SEQUENCES
                    #
                    try:
                        observed_surrogate_sequence_encoded = crispr_sequence_encoding.encode_DNA_base_vectorized(crispr_sequence_encoding.numpify_string_vectorized(observed_guide_reporter_sequence["surrogate"]))  # Encode the observed protospacer
                    except Exception as e:
                        print(e)
                        raise Exception(f"Error encoding surrogate sequence: {observed_guide_reporter_sequence['surrogate']}")
                    observed_surrogate_sequence_dists = crispr_sequence_encoding.retrieve_hamming_distance_whitelist(observed_surrogate_sequence_encoded, encoded_whitelist_surrogate_sequences_series) # Hamming distance among  all whitelist surrogates
                    observed_surrogate_sequence_dists_min = observed_surrogate_sequence_dists.min()
                    surrogate_hamming_threshold_met = (observed_surrogate_sequence_dists_min < surrogate_hamming_threshold)

                    surrogate_match_indices = np.where(observed_surrogate_sequence_dists == observed_surrogate_sequence_dists_min)[0]
                    whitelist_guide_reporter_df_hamming_surrogate_match = whitelist_guide_reporter_df.iloc[surrogate_match_indices] # Get all whitelisted guides with the minimum hamming distance (could be multiple)
                    if surrogate_hamming_threshold_met: # IF SURROGATE MATCHED

                        # See if there are identical matches between protospacer-only matches and surrogate-only matches
                        whitelist_guide_reporter_df_hamming_surrogate_match_protospacer_match = pd.merge(whitelist_guide_reporter_df_hamming_surrogate_match, whitelist_guide_reporter_df_hamming_protospacer_match, how='inner')
                        complete_match_result.protospacer_mismatch_surrogate_match = SurrogateProtospacerMismatchSingleInferenceMatchResult(value=SurrogateProtospacerMismatchSingleInferenceMatchResultValue(
                            mismatched=whitelist_guide_reporter_df_hamming_surrogate_match_protospacer_match.empty, 
                            surrogate_matches=whitelist_guide_reporter_df_hamming_surrogate_match, 
                            protospacer_matches=whitelist_guide_reporter_df_hamming_protospacer_match,
                            protospacer_surrogate_matches=whitelist_guide_reporter_df_hamming_surrogate_match_protospacer_match
                        ))
                    else: # NO SURROGATE MATCH, ERROR
                        complete_match_result.protospacer_mismatch_surrogate_match = SurrogateProtospacerMismatchSingleInferenceMatchResult(
                            error=SurrogateHammingThresholdGuideCountError(
                                hamming_min=observed_surrogate_sequence_dists_min,
                                hamming_threshold=surrogate_hamming_threshold,
                                original_df=whitelist_guide_reporter_df,
                                hamming_min_match_df=whitelist_guide_reporter_df_hamming_surrogate_match,
                                barcode_subsetted=False
                            ))
                else: # SURROGATE PARSING ISSUE, ERROR
                    complete_match_result.protospacer_mismatch_surrogate_match_barcode_match = SurrogateProtospacerMismatchSingleInferenceMatchResult(error=surrogate_error_result)
                    complete_match_result.protospacer_mismatch_surrogate_match = SurrogateProtospacerMismatchSingleInferenceMatchResult(error=surrogate_error_result)
                
        else: # NO PROTOSPACER MATCH, ERROR
            error_result = ProtospacerHammingThresholdGuideCountError(
                hamming_min=observed_protospacer_sequence_dists_min,
                hamming_threshold=protospacer_hamming_threshold,
                original_df=whitelist_guide_reporter_df,
                hamming_min_match_df=whitelist_guide_reporter_df_hamming_protospacer_match,
                barcode_subsetted=False
            )
            complete_match_result.protospacer_match = MatchSetSingleInferenceMatchResult(error=error_result)
            complete_match_result.protospacer_match_surrogate_match_barcode_match = MatchSetSingleInferenceMatchResult(error=error_result)
            complete_match_result.protospacer_match_surrogate_match = MatchSetSingleInferenceMatchResult(error=error_result)
            complete_match_result.protospacer_match_barcode_match = MatchSetSingleInferenceMatchResult(error=error_result)
            complete_match_result.protospacer_mismatch_surrogate_match_barcode_match = SurrogateProtospacerMismatchSingleInferenceMatchResult(error=error_result)
            complete_match_result.protospacer_mismatch_surrogate_match = SurrogateProtospacerMismatchSingleInferenceMatchResult(error=error_result)
    else: # PARSED PROTOSPACER HAS SOME UNEXPECTED ISSUE, ERROR
        complete_match_result.protospacer_match = MatchSetSingleInferenceMatchResult(error=protospacer_error_result)
        complete_match_result.protospacer_match_surrogate_match_barcode_match = MatchSetSingleInferenceMatchResult(error=protospacer_error_result)
        complete_match_result.protospacer_match_surrogate_match = MatchSetSingleInferenceMatchResult(error=protospacer_error_result)
        complete_match_result.protospacer_match_barcode_match = MatchSetSingleInferenceMatchResult(error=protospacer_error_result)
        complete_match_result.protospacer_mismatch_surrogate_match_barcode_match = SurrogateProtospacerMismatchSingleInferenceMatchResult(error=protospacer_error_result)
        complete_match_result.protospacer_mismatch_surrogate_match = SurrogateProtospacerMismatchSingleInferenceMatchResult(error=protospacer_error_result)
        
    
    return complete_match_result



'''
    This performs simulation of determining how many mutations it takes for a guide to be ambigiously mapped based on hamming distance.

    This is performed on *sample_count* number of randomly selected guides, and the distribution of mutation count is determined. The 5% quantile is used as the threshold.
    
    This is useful in determing the ideal hamming distance threshold specific to a guide library
'''
@typechecked
def determine_hamming_threshold(whitelist_sequences_series: Union[List[str],pd.Series], encoded_whitelist_sequences_series, sample_count: int = 100, quantile: float = 0.05) -> int:
    #encoded_whitelist_guide_sequences_series = encode_guide_series(guide_sequences_series)
    
    mutation_count_until_nonunique = []
    for i in range(sample_count):
        # Sample a guide from whitelist
        sampled_guide = whitelist_sequences_series.sample()[0]
        
        # Generate order of base positions to mutate sequentially.
        guide_position_list = list(range(len(sampled_guide)))
        random.shuffle(guide_position_list)
        
        # Create temporary variable to represent mutated guide
        current_guide_sequence = sampled_guide
        
        # Iterate through positions to mutate
        for iteration, position in enumerate(guide_position_list):
            
            # Mutate the guide
            current_guide_sequence_separated = list(current_guide_sequence)
            guide_position_nt = current_guide_sequence_separated[position]
            nt_list = ["A", "C", "T", "G"]
            nt_list.remove(guide_position_nt.upper())
            new_nt = random.sample(nt_list, 1)[0]
            current_guide_sequence_separated[position] = new_nt
            current_guide_sequence = "".join(current_guide_sequence_separated)
            try:
                current_guide_sequence_encoded = crispr_sequence_encoding.encode_DNA_base_vectorized(crispr_sequence_encoding.numpify_string_vectorized(current_guide_sequence)) 
            except Exception as e:
                print(e)
                raise Exception(f"Error encoding surrogate sequence: {current_guide_sequence}")
            # Calculate the hamming distance of the mutated guide to all other guides
            hamming_distances =  crispr_sequence_encoding.retrieve_hamming_distance_whitelist(current_guide_sequence_encoded, encoded_whitelist_sequences_series)
            if len(np.where(hamming_distances == hamming_distances.min())[0]) > 1: # TODO (3/30/23): Can also add to the conditional whether the minimum hamming distance guide is still the original guide - but it is probably rare cases where this would not be the case, so not too important to implement
                mutation_count_until_nonunique.append(iteration+1)
                break   
    mutation_count_until_nonunique = pd.Series(mutation_count_until_nonunique)
    
    # From the mutation count distribution, calculate the threshold based on the provided quantile.
    hamming_distance_threshold = mutation_count_until_nonunique.quantile(quantile)
    return int(hamming_distance_threshold) + 1 # Take ceil of result (to be conservative)




from dataclasses import dataclass
from typing import Union, List, Mapping, Tuple, Optional, Any, DefaultDict, Dict
from typing import Counter as CounterType

from collections import Counter, defaultdict
import pandas as pd

from .error_models import GuideCountError
from .quality_control_models import QualityControlResult

from typing import Union, List, Mapping, Tuple, Optional, Any, DefaultDict, Dict
from typing import Counter as CounterType
import pandas as pd

# Provides a mapping between the whitelist sequence and the list of observed sequences. Each observed sequence has the corresponding count Union[int, Dict [ str, int ]] (either int if no UMI, or Dict[str, int] for UMI-collapsed and non-collpased count)
WhitelistReporterObservedSequenceMapping = DefaultDict[  Tuple[str, Optional[str], Optional[str]],    List[   Tuple[ Tuple[str, Optional[str], Optional[str]], Union[int, Dict [ str, int ]], int, List[Tuple[str, Optional[str], Optional[str]]] ]    ]]

# Sequence Count Result Objects
ProtospacerCounter = CounterType[str]
ProtospacerSurrogateCounter = CounterType[Tuple[str, str]]
ProtospacerBarcodeCounter = CounterType[Tuple[str, str]]
ProtospacerSurrogateBarcodeCounter = CounterType[Tuple[str, str, str]]

ProtospacerDictUMICounter = DefaultDict[str, CounterType[str]]
ProtospacerSurrogateDictUMICounter = DefaultDict[Tuple[str, str], CounterType[str]]
ProtospacerBarcodeDictUMICounter = DefaultDict[Tuple[str, str], CounterType[str]]
ProtospacerSurrogateBarcodeDictUMICounter = DefaultDict[Tuple[str, str, str], CounterType[str]]

GeneralGuideCountType = Union[ProtospacerCounter, 
                       ProtospacerSurrogateCounter, 
                       ProtospacerBarcodeCounter, 
                       ProtospacerSurrogateBarcodeCounter,
                       ProtospacerDictUMICounter,
                       ProtospacerSurrogateDictUMICounter,
                       ProtospacerBarcodeDictUMICounter, 
                       ProtospacerSurrogateBarcodeDictUMICounter]

# Mapping Count Dict Object

ProtospacerSurrogateBarcodeMatchCountDict = DefaultDict[Tuple[str, str, str], Union[int, float]]
ProtospacerSurrogateMatchCountDict = DefaultDict[Tuple[str, str], Union[int, float]]
ProtospacerBarcodeMatchCountDict = DefaultDict[Tuple[str, str], Union[int, float]]
ProtospacerMatchCountDict = DefaultDict[str, Union[int, float]]
GeneralMatchCountDict = Union[ProtospacerSurrogateBarcodeMatchCountDict, 
      ProtospacerSurrogateMatchCountDict, 
      ProtospacerBarcodeMatchCountDict, 
      ProtospacerMatchCountDict]

ProtospacerSurrogateBarcodeMismatchCountDict = DefaultDict[Tuple[Tuple[str, str, str], Tuple[str, str, str]], Union[int, float]]
ProtospacerSurrogateMismatchCountDict = DefaultDict[Tuple[Tuple[str, str], Tuple[str, str]], Union[int, float]]
ProtospacerBarcodeMismatchCountDict = DefaultDict[Tuple[Tuple[str, str], Tuple[str, str]], Union[int, float]]
ProtospacerMismatchCountDict = DefaultDict[Tuple[str, str], Union[int, float]]

GeneralMismatchCountDict = Union[ProtospacerSurrogateBarcodeMismatchCountDict, 
      ProtospacerSurrogateMismatchCountDict, 
      ProtospacerBarcodeMismatchCountDict, 
      ProtospacerMismatchCountDict]

# Allele nested dict Object (Key of first dict is inferred, key of second dict is observed, value of second dict is count)
ProtospacerSurrogateBarcodeAlleleDict = DefaultDict[Tuple[str, str, str], DefaultDict[Tuple[str, str, str], Union[int, float]]]
ProtospacerSurrogateAlleleDict = DefaultDict[Tuple[str, str], DefaultDict[Tuple[str, str], Union[int, float]]]
ProtospacerBarcodeAlleleDict = DefaultDict[Tuple[str, str], DefaultDict[Tuple[str, str], Union[int, float]]]
ProtospacerAlleleDict = DefaultDict[str, DefaultDict[str, Union[int, float]]]

GeneralAlleleDict = Union[ProtospacerSurrogateBarcodeAlleleDict,
      ProtospacerSurrogateAlleleDict,
      ProtospacerBarcodeAlleleDict,
      ProtospacerAlleleDict]




# Allele Count Series Dict
ProtospacerSurrogateBarcodeAlleleCountSeriesDict = DefaultDict[Tuple[str, str, str], pd.Series]
ProtospacerSurrogateAlleleCountSeriesDict = DefaultDict[Tuple[str, str], pd.Series]
ProtospacerBarcodeAlleleCountSeriesDict = DefaultDict[Tuple[str, str], pd.Series]
ProtospacerAlleleCountSeriesDict = DefaultDict[str, pd.Series]

GeneralAlleleCountSeriesDict = Union[ProtospacerSurrogateBarcodeAlleleCountSeriesDict,
                                     ProtospacerSurrogateAlleleCountSeriesDict,
                                     ProtospacerBarcodeAlleleCountSeriesDict,
                                     ProtospacerAlleleCountSeriesDict]







@dataclass
class SingleInferenceMatchResultValue:
    pass

@dataclass
class MatchSetSingleInferenceMatchResultValue(SingleInferenceMatchResultValue):
    matches: Optional[pd.DataFrame] = None

@dataclass
class SurrogateProtospacerMismatchSingleInferenceMatchResultValue(SingleInferenceMatchResultValue):
    mismatched: bool
    surrogate_matches: pd.DataFrame
    protospacer_matches: pd.DataFrame 
    protospacer_surrogate_matches: pd.DataFrame

@dataclass
class SingleInferenceMatchResult:
    error: Optional[GuideCountError] = None
    value: Optional[SingleInferenceMatchResultValue] = None

@dataclass
class MatchSetSingleInferenceMatchResult(SingleInferenceMatchResult):
    value: Optional[MatchSetSingleInferenceMatchResultValue] = None

@dataclass
class SurrogateProtospacerMismatchSingleInferenceMatchResult(SingleInferenceMatchResult):
    value: Optional[SurrogateProtospacerMismatchSingleInferenceMatchResultValue] = None

@dataclass
class CompleteInferenceMatchResult:
    protospacer_match: Optional[MatchSetSingleInferenceMatchResult] = None
    protospacer_match_surrogate_match_barcode_match: Optional[MatchSetSingleInferenceMatchResult] = None
    protospacer_match_surrogate_match: Optional[MatchSetSingleInferenceMatchResult] = None
    protospacer_match_barcode_match: Optional[MatchSetSingleInferenceMatchResult] =None
    protospacer_mismatch_surrogate_match_barcode_match: Optional[SurrogateProtospacerMismatchSingleInferenceMatchResult] = None
    protospacer_mismatch_surrogate_match: Optional[SurrogateProtospacerMismatchSingleInferenceMatchResult] = None

@dataclass
class InferenceResult:
    observed_value: Union[int, CounterType[Optional[str]]]
    inferred_value: CompleteInferenceMatchResult


# Inference Result Object
ProtospacerSurrogateBarcodeMappingInferenceDict = DefaultDict[Tuple[str,str,str], Dict[Tuple[str,str,str], InferenceResult]]
ProtospacerSurrogateMappingInferenceDict = DefaultDict[Tuple[str,str], Dict[Tuple[str,str], InferenceResult]]
ProtospacerBarcodeMappingInferenceDict = DefaultDict[Tuple[str,str], Dict[Tuple[str,str], InferenceResult]]
ProtospacerMappingInferenceDict = DefaultDict[str, Dict[str, InferenceResult]]

GeneralMappingInferenceDict = Union[ProtospacerSurrogateBarcodeMappingInferenceDict,
      ProtospacerSurrogateMappingInferenceDict,
      ProtospacerBarcodeMappingInferenceDict,
      ProtospacerMappingInferenceDict]


@dataclass
class MatchSetWhitelistReporterCounterSeriesResults:
    ambiguous_ignored_umi_noncollapsed_counterseries : Optional[pd.Series] = None
    ambiguous_ignored_umi_collapsed_counterseries : Optional[pd.Series] = None
    ambiguous_ignored_counterseries : Optional[pd.Series] = None

    ambiguous_accepted_umi_noncollapsed_counterseries : Optional[pd.Series] = None
    ambiguous_accepted_umi_collapsed_counterseries : Optional[pd.Series] = None
    ambiguous_accepted_counterseries : Optional[pd.Series] = None

    ambiguous_spread_umi_noncollapsed_counterseries : Optional[pd.Series] = None
    ambiguous_spread_umi_collapsed_counterseries : Optional[pd.Series] = None
    ambiguous_spread_counterseries : Optional[pd.Series] = None

    # This ensures that any empty series are kept at None
    def __setattr__(self, name, value):
        if isinstance(value, pd.Series):
            if value.sum() == 0: # If sum of pandas series is 0, just return None
                super().__setattr__(name, None)
            else:
                super().__setattr__(name, value)
        else:
            # Set the attribute
            super().__setattr__(name, value)

@dataclass
class SurrogateProtospacerMismatchSetWhitelistReporterCounterSeriesResults:
    
    # MATCH counters
    ambiguous_ignored_umi_noncollapsed_match_counterseries : Optional[pd.Series]  = None
    ambiguous_ignored_umi_collapsed_match_counterseries : Optional[pd.Series]  = None
    ambiguous_ignored_match_counterseries : Optional[pd.Series] = None

    ambiguous_accepted_umi_noncollapsed_match_counterseries : Optional[pd.Series] = None
    ambiguous_accepted_umi_collapsed_match_counterseries : Optional[pd.Series] = None
    ambiguous_accepted_match_counterseries : Optional[pd.Series] = None

    ambiguous_spread_umi_noncollapsed_match_counterseries : Optional[pd.Series] = None
    ambiguous_spread_umi_collapsed_match_counterseries : Optional[pd.Series] = None
    ambiguous_spread_match_counterseries : Optional[pd.Series] = None

    # MISMATCH counters (keys are PAIRS of indices, representing the protospacer and surrogate match separately), [str, Optional[str], Optional[str], str, Optional[str], Optional[str]], first three are Protospacer_Match, last three is Surrogate_Match
    ambiguous_ignored_umi_noncollapsed_mismatch_counterseries : Optional[pd.Series] = None
    ambiguous_ignored_umi_collapsed_mismatch_counterseries : Optional[pd.Series] = None
    ambiguous_ignored_mismatch_counterseries : Optional[pd.Series] = None

    ambiguous_accepted_umi_noncollapsed_mismatch_counterseries : Optional[pd.Series] = None
    ambiguous_accepted_umi_collapsed_mismatch_counterseries : Optional[pd.Series] = None
    ambiguous_accepted_mismatch_counterseries : Optional[pd.Series] = None

    ambiguous_spread_umi_noncollapsed_mismatch_counterseries : Optional[pd.Series] = None
    ambiguous_spread_umi_collapsed_mismatch_counterseries : Optional[pd.Series] = None
    ambiguous_spread_mismatch_counterseries : Optional[pd.Series] = None

    # This ensures that any empty series are kept at None
    def __setattr__(self, name, value):
        if isinstance(value, pd.Series):
            if value.sum() == 0: # If sum of pandas series is 0, just return None
                super().__setattr__(name, None)
            else:
                super().__setattr__(name, value)
        else:
            # Set the attribute
            super().__setattr__(name, value)

@dataclass
class AllMatchSetWhitelistReporterCounterSeriesResults:
    protospacer_match: Optional[MatchSetWhitelistReporterCounterSeriesResults] = None
    protospacer_match_surrogate_match_barcode_match: Optional[MatchSetWhitelistReporterCounterSeriesResults] = None
    protospacer_match_surrogate_match: Optional[MatchSetWhitelistReporterCounterSeriesResults] = None
    protospacer_match_barcode_match: Optional[MatchSetWhitelistReporterCounterSeriesResults] = None
    protospacer_mismatch_surrogate_match_barcode_match: Optional[SurrogateProtospacerMismatchSetWhitelistReporterCounterSeriesResults] = None
    protospacer_mismatch_surrogate_match: Optional[SurrogateProtospacerMismatchSetWhitelistReporterCounterSeriesResults] = None

@dataclass
class CountInput:
    whitelist_guide_reporter_df: pd.DataFrame
    contains_surrogate:bool
    contains_barcode:bool
    contains_umi:bool
    protospacer_hamming_threshold_strict: Optional[int]
    surrogate_hamming_threshold_strict: Optional[int]
    barcode_hamming_threshold_strict: Optional[int]

@dataclass
class WhitelistReporterCountsResult:
    all_match_set_whitelist_reporter_counter_series_results: AllMatchSetWhitelistReporterCounterSeriesResults
    observed_guide_reporter_umi_counts_inferred: GeneralMappingInferenceDict
    quality_control_result: QualityControlResult
    count_input: CountInput



#
# Types
#





import pandas as pd
import numpy as np
import Bio
from Bio import SeqIO
from Bio.Seq import Seq
from pandarallel import pandarallel
import matplotlib.pyplot as plt
from collections import Counter, defaultdict
from typeguard import typechecked
from datetime import date
import re
from multiprocessing import Pool
from functools import partial
from itertools import repeat
import gzip
import random
from enum import Enum
from typing import Callable
from typing import Union, List, Mapping, Tuple, Optional, Any, DefaultDict
from typing import Counter as CounterType
from concurrent.futures import ProcessPoolExecutor
from dataclasses import dataclass

from ..processing import crispr_sequence_encoding
from ..processing import crispr_guide_counting
from ..processing import crispr_sequence_encoding
from ..parsing import reporter_umitools_fastq_parsing, reporter_standard_fastq_parsing
from ..models.mapping_models import WhitelistReporterCountsResult


@typechecked
def get_whitelist_reporter_counts_from_umitools_output(whitelist_guide_reporter_df: pd.DataFrame, 
                                                       fastq_r1_fn: str, 
                                                       fastq_r2_fn: str, 
                                                       barcode_pattern_regex: Optional[str] = None, 
                                                       umi_pattern_regex: Optional[str] = None, 
                                                       revcomp_protospacer: bool = False, 
                                                       revcomp_surrogate: bool = True, 
                                                       revcomp_barcode: bool = True, 
                                                       surrogate_hamming_threshold_strict: Optional[int] = 10, 
                                                       barcode_hamming_threshold_strict: Optional[int] = 2, 
                                                       protospacer_hamming_threshold_strict: Optional[int] = 7, 
                                                       cores: int=1) -> WhitelistReporterCountsResult:
    #
    # Get counts of observed FASTQ sequences
    #
    observed_guide_reporter_umi_counts = reporter_umitools_fastq_parsing.get_umitools_observed_sequence_counts(r1_protospacer_fastq_file=fastq_r1_fn, r2_surrogate_fastq_file=fastq_r2_fn, barcode_pattern_regex=barcode_pattern_regex, umi_pattern_regex=umi_pattern_regex, revcomp_protospacer = revcomp_protospacer, revcomp_surrogate = revcomp_surrogate, revcomp_barcode = revcomp_barcode)

    #   
    # Map the observed sequences to the true whitelist sequence
    #
    get_whitelist_reporter_counts_with_umi_PARTIAL = partial(
        crispr_guide_counting.get_whitelist_reporter_counts_with_umi,
        observed_guide_reporter_umi_counts=observed_guide_reporter_umi_counts,
        whitelist_guide_reporter_df=whitelist_guide_reporter_df,
        protospacer_hamming_threshold_strict=protospacer_hamming_threshold_strict,
        surrogate_hamming_threshold_strict=surrogate_hamming_threshold_strict,
        barcode_hamming_threshold_strict=barcode_hamming_threshold_strict,
        cores=cores
    )
    
    if fastq_r2_fn is None: # ONLY R1
        if barcode_pattern_regex is None: # ONLY R1; NO BARCODE
            if umi_pattern_regex is None: # ONLY R1; NO BARCODE; NO UMI
                # These lines are really added just to document the type... observed_guide_reporter_umi_counts is already added to the partial above anyways
                observed_guide_reporter_umi_counts: CounterType[str] = observed_guide_reporter_umi_counts
                return get_whitelist_reporter_counts_with_umi_PARTIAL(contains_surrogate=False, contains_barcode=False, contains_umi = False)
            else: # ONLY R1; NO BARCODE; YES UMI
                observed_guide_reporter_umi_counts: DefaultDict[str, CounterType[str]] = observed_guide_reporter_umi_counts
                return get_whitelist_reporter_counts_with_umi_PARTIAL(contains_surrogate=False, contains_barcode=False, contains_umi = True)
        else: # ONLY R1; YES BARCODE
            if umi_pattern_regex is None: # ONLY R1; YES BARCODE; NO UMI
                observed_guide_reporter_umi_counts: CounterType[Tuple[str, str]] = observed_guide_reporter_umi_counts
                return get_whitelist_reporter_counts_with_umi_PARTIAL(contains_surrogate=False, contains_barcode=True, contains_umi = False)
            else: # ONLY R1; YES BARCODE; YES UMI
                observed_guide_reporter_umi_counts: DefaultDict[Tuple[str, str], CounterType[str]] = observed_guide_reporter_umi_counts
                return get_whitelist_reporter_counts_with_umi_PARTIAL(contains_surrogate=False, contains_barcode=True, contains_umi = True)
    else: # YES R2
        if barcode_pattern_regex is None: # YES R2; NO BARCODE
            if umi_pattern_regex is None: # YES R2; NO BARCODE; NO UMI
                observed_guide_reporter_umi_counts: CounterType[Tuple[str, str]] = observed_guide_reporter_umi_counts
                return get_whitelist_reporter_counts_with_umi_PARTIAL(contains_surrogate=True, contains_barcode=False, contains_umi = False)
            else: # YES R2; NO BARCODE; YES UMI
                observed_guide_reporter_umi_counts: DefaultDict[Tuple[str, str], CounterType[str]] = observed_guide_reporter_umi_counts
                return get_whitelist_reporter_counts_with_umi_PARTIAL(contains_surrogate=True, contains_barcode=False, contains_umi = True)
        else: # YES R2; YES BARCODE
            if umi_pattern_regex is None: # YES R2; YES BARCODE; NO UMI
                observed_guide_reporter_umi_counts: CounterType[Tuple[str, str, str]] = observed_guide_reporter_umi_counts
                return get_whitelist_reporter_counts_with_umi_PARTIAL(contains_surrogate=True, contains_barcode=True, contains_umi = False)
            else: # YES R2; YES BARCODE; YES UMI
                observed_guide_reporter_umi_counts: DefaultDict[Tuple[str, str, str], CounterType[str]] = observed_guide_reporter_umi_counts
                return get_whitelist_reporter_counts_with_umi_PARTIAL(contains_surrogate=True, contains_barcode=True, contains_umi = True)


# TODO 11/20/2024: Implement UMI parsing via sequence or via R1 header
#@typechecked
def get_whitelist_reporter_counts_from_fastq(whitelist_guide_reporter_df: pd.DataFrame, 
                                                       fastq_r1_fn: str, 
                                                       fastq_r2_fn: Optional[str] = None, 
                                                       
                                                       protospacer_pattern_regex: Optional[str] = None,
                                                       surrogate_pattern_regex: Optional[str] = None,
                                                       barcode_pattern_regex: Optional[str] = None,
                                                       umi_pattern_regex: Optional[str] = None,

                                                       protospacer_left_flank:Optional[str] = None,
                                                       protospacer_right_flank:Optional[str] = None,
                                                       protospacer_start_position:Optional[int] = None,
                                                       protospacer_end_position:Optional[int] = None,
                                                       protospacer_length: Optional[int] = None,

                                                       surrogate_left_flank:Optional[str] = None,
                                                       surrogate_right_flank:Optional[str] = None,
                                                       surrogate_start_position:Optional[int] = None,
                                                       surrogate_end_position:Optional[int] = None,
                                                       surrogate_length: Optional[int] = None,

                                                       barcode_left_flank:Optional[str] = None,
                                                       barcode_right_flank:Optional[str] = None,
                                                       barcode_start_position:Optional[int] = None,
                                                       barcode_end_position:Optional[int] = None,
                                                       barcode_length: Optional[int] = None,

                                                       umi_left_flank:Optional[str] = None,
                                                       umi_right_flank:Optional[str] = None,
                                                       umi_start_position:Optional[int] = None,
                                                       umi_end_position:Optional[int] = None,
                                                       umi_length: Optional[int] = None,

                                                       is_protospacer_r1: Optional[bool] = None, 
                                                       is_surrogate_r1: Optional[bool] = None, 
                                                       is_barcode_r1: Optional[bool] = None,
                                                       is_umi_r1: Optional[bool] = None,
                                                       is_protospacer_header: Optional[bool] = None, 
                                                       is_surrogate_header: Optional[bool] = None, 
                                                       is_barcode_header: Optional[bool] = None,
                                                       is_umi_header: Optional[bool] = None,
                                                    
                                                       revcomp_protospacer: Optional[bool] = None, 
                                                       revcomp_surrogate: Optional[bool] = None, 
                                                       revcomp_barcode: Optional[bool] = None, 
                                                       revcomp_umi: Optional[bool] = None,
                                                        
                                                       surrogate_hamming_threshold_strict: Optional[int] = None, 
                                                       barcode_hamming_threshold_strict: Optional[int] = None, 
                                                       protospacer_hamming_threshold_strict: Optional[int] = None, 
                                                       cores: int=1) -> WhitelistReporterCountsResult:
    # Input parameter validation checks

    protospacer_pattern_regex = None if ((protospacer_pattern_regex is not None) and  (protospacer_pattern_regex.strip() == "")) else protospacer_pattern_regex
    surrogate_pattern_regex = None if ((surrogate_pattern_regex is not None) and (surrogate_pattern_regex.strip() == "")) else surrogate_pattern_regex
    barcode_pattern_regex = None if ((barcode_pattern_regex is not None) and  (barcode_pattern_regex.strip() == "")) else barcode_pattern_regex
    umi_pattern_regex = None if ((umi_pattern_regex is not None) and (umi_pattern_regex.strip() == "")) else umi_pattern_regex

    contains_surrogate = (surrogate_pattern_regex is not None) or (surrogate_left_flank is not None) or (surrogate_right_flank is not None) or (surrogate_start_position is not None) or (surrogate_end_position is not None) or (surrogate_length is not None)
    contains_barcode = (barcode_pattern_regex is not None) or (barcode_left_flank is not None) or (barcode_right_flank is not None) or (barcode_start_position is not None) or (barcode_end_position is not None) or (barcode_length is not None)
    contains_umi = (umi_pattern_regex is not None) or (umi_left_flank is not None) or (umi_right_flank is not None) or (umi_start_position is not None) or (barcode_end_position is not None) or (barcode_length is not None)
    

    #
    # Get counts of observed FASTQ sequences
    #
    observed_guide_reporter_umi_counts = reporter_standard_fastq_parsing.get_standard_observed_sequence_counts(fastq_r1_fn=fastq_r1_fn, 
                                            fastq_r2_fn=fastq_r2_fn, 
                                            
                                            protospacer_pattern_regex=protospacer_pattern_regex,
                                            surrogate_pattern_regex=surrogate_pattern_regex,
                                            barcode_pattern_regex=barcode_pattern_regex,
                                            umi_pattern_regex=umi_pattern_regex,

                                            protospacer_left_flank=protospacer_left_flank,
                                            protospacer_right_flank=protospacer_right_flank,
                                            protospacer_start_position=protospacer_start_position,
                                            protospacer_end_position=protospacer_end_position,
                                            protospacer_length=protospacer_length,

                                            surrogate_left_flank=surrogate_left_flank,
                                            surrogate_right_flank=surrogate_right_flank,
                                            surrogate_start_position=surrogate_start_position,
                                            surrogate_end_position=surrogate_end_position,
                                            surrogate_length=surrogate_length,

                                            barcode_left_flank=barcode_left_flank,
                                            barcode_right_flank=barcode_right_flank,
                                            barcode_start_position=barcode_start_position,
                                            barcode_end_position=barcode_end_position,
                                            barcode_length=barcode_length,

                                            umi_left_flank=umi_left_flank,
                                            umi_right_flank=umi_right_flank,
                                            umi_start_position=umi_start_position,
                                            umi_end_position=umi_end_position,
                                            umi_length=umi_length,

                                            is_protospacer_r1=is_protospacer_r1, 
                                            is_surrogate_r1=is_surrogate_r1, 
                                            is_barcode_r1=is_barcode_r1,
                                            is_umi_r1=is_umi_r1,
                                            is_protospacer_header=is_protospacer_header, 
                                            is_surrogate_header=is_surrogate_header, 
                                            is_barcode_header=is_barcode_header,
                                            is_umi_header=is_umi_header,
                                        
                                            revcomp_protospacer=revcomp_protospacer, 
                                            revcomp_surrogate=revcomp_surrogate, 
                                            revcomp_barcode=revcomp_barcode,
                                            revcomp_umi=revcomp_umi,
                                            
                                            contains_surrogate=contains_surrogate,
                                            contains_barcode=contains_barcode,
                                            contains_umi=contains_umi)

    #   
    # Map the observed sequences to the true whitelist sequence
    #
    get_whitelist_reporter_counts_with_umi_PARTIAL = partial(
        crispr_guide_counting.get_whitelist_reporter_counts_with_umi,
        observed_guide_reporter_umi_counts=observed_guide_reporter_umi_counts,
        whitelist_guide_reporter_df=whitelist_guide_reporter_df,
        protospacer_hamming_threshold_strict=protospacer_hamming_threshold_strict,
        surrogate_hamming_threshold_strict=surrogate_hamming_threshold_strict,
        barcode_hamming_threshold_strict=barcode_hamming_threshold_strict,
        cores=cores
    )

    return get_whitelist_reporter_counts_with_umi_PARTIAL(contains_surrogate=contains_surrogate, contains_barcode=contains_barcode, contains_umi = contains_umi)
                




from typeguard import typechecked
from typing import Union
from typing import Callable
from typing import Counter as CounterType
from functools import partial
from collections import Counter

# DEPRECATED CLASS

'''
    Strategies to parse read
'''
@typechecked
def parse_read_positional(read_sequence: str, position_start: int, position_end: int) -> str:  
    return read_sequence[position_start:position_end]

@typechecked
def parse_read_left_flank(read_sequence: str, left_flank:str, guide_sequence_length:int) -> str: 
    position_start = read_sequence.find(left_flank) + len(left_flank)
    return read_sequence[position_start:position_start+guide_sequence_length]

@typechecked
def parse_read_right_flank(read_sequence: str, right_flank:str, guide_sequence_length:int) -> str:
    position_end = read_sequence.find(right_flank)
    return read_sequence[position_end-guide_sequence_length:position_end]


'''
    Extract the guide sequence from the read provided
'''
@typechecked
def parse_guide_sequence(read_sequence: str, parser_function: Callable) -> str:
    read_guide_sequence = parser_function(read_sequence)
    return read_guide_sequence




@typechecked
def get_raw_fastq_observed_sequence_counts(fastq_file: str, parse_left_flank: bool = True, parse_flank_sequence: Union[None, str] = None, cores: int=1) -> CounterType[str]:
    parse_guide_sequence_p = None
    if parse_left_flank:
        if parse_flank_sequence is None:
            print("No flank sequence passed. Setting left-flank default sequence to CACCG assuming U6 G+N20 guide")
            flank_sequence = "CACCG"
        else:
            flank_sequence = parse_flank_sequence

        parse_read_left_flank_p = partial(parse_read_left_flank, left_flank=flank_sequence, guide_sequence_length=20)
        parse_guide_sequence_p = partial(parse_guide_sequence, parser_function=parse_read_left_flank_p)
    else:
        if parse_flank_sequence is None:
            print("No flank sequence passed. Setting right-flank default sequence to GTTTT. If you are using sgRNA(F+E) design, flank sequence may need to be changed")
            flank_sequence = "GTTTT"
        else:
            flank_sequence = parse_flank_sequence
        
        parse_read_right_flank_p = partial(parse_read_right_flank, right_flank=flank_sequence, guide_sequence_length=20)
        parse_guide_sequence_p = partial(parse_guide_sequence, parser_function=parse_read_right_flank_p)

    # Looks for a flanking sequences (CACCG, or GTTTT) in the read, then extracts the 20nt guide sequence


    import gzip

    def parse_fastq_guide_sequences(file_handler):
        fastq_guide_sequences = []
        for line_number, line in enumerate(file):
            if line_number % 4 == 1:
                fastq_guide_sequences.append(parse_guide_sequence_p(line.strip()))
        
        # Count the unique observed sequences
        sequence_counter = Counter(fastq_guide_sequences)
        return sequence_counter
    
    if fastq_file.endswith('.gz'):
            print(f"Opening FASTQ.gz file with gzip, filename={fastq_file}")
            with gzip.open(fastq_file, "rt", encoding="utf-8") as file:
                return parse_fastq_guide_sequences(file)
    else:
        with open(fastq_file, "r") as file:
            return parse_fastq_guide_sequences(file)


#!/usr/bin/env python

import pickle
from datetime import date
from os import listdir
from os.path import isfile, join

def save_or_load_pickle(directory, label, py_object = None, date_string = None):
    '''Save a pickle for caching that is notated by the date'''
    
    if date_string == None:
        today = date.today()
        date_string = str(today.year) + ("0" + str(today.month) if today.month < 10 else str(today.month)) + str(today.day)
    
    filename = directory + label + "_" + date_string + '.pickle'
    print(filename)
    if py_object is None:
        with open(filename, 'rb') as handle:
            py_object = pickle.load(handle)
            return py_object
    else:
        with open(filename, 'wb') as handle:
            pickle.dump(py_object, handle, protocol=pickle.HIGHEST_PROTOCOL)

def display_all_pickle_versions(directory, label):
    '''Retrieve all pickles with a label, specifically to identify versions available'''
    return [f for f in listdir(directory) if isfile(join(directory, f)) and label == f[:len(label)]]



from .crispr_editing_processing import *
from .crispr_guide_counting import *
from .crispr_guide_inference import *
from .crispr_sequence_encoding import *
from .crispr_count_processing import helper_get_observed_values_given_whitelist_value


from collections import Counter, defaultdict
from dataclasses import dataclass
from typing import Optional, DefaultDict, Union, Tuple, List
import pandas as pd
from typing import Counter as CounterType
from ..models.mapping_models import GeneralMappingInferenceDict, GeneralAlleleDict, GeneralAlleleCountSeriesDict
from ..models.mapping_models import CompleteInferenceMatchResult, MatchSetSingleInferenceMatchResult, InferenceResult
from ..models.editing_models import (MatchSetWhitelistReporterObservedSequenceCounterSeriesResults, 
                                     MatchSetWhitelistReporterObservedSequenceMutationProfiles, 
                                     LinkedMutationCounters, 
                                     ObservedSequenceMutationProfile)

# Function that has conditional on whether a match result is not an errror
def check_match_result_non_error(match_result):
    return False if match_result is None else match_result.error is None # If match_result is None, treat as error. If match_result is not None, but error is None, then non_error

# Filter dict with observed sequence inference results for only those that do not contain any mapping errors
def get_non_error_dict(observed_guide_reporter_umi_counts_inferred: GeneralMappingInferenceDict, attribute_name: str) -> MatchSetWhitelistReporterObservedSequenceCounterSeriesResults:
    return {observed_guide_reporter_key: observed_guide_reporter_umi_counts_inferred_value for observed_guide_reporter_key, observed_guide_reporter_umi_counts_inferred_value in observed_guide_reporter_umi_counts_inferred.items() if check_match_result_non_error(getattr(observed_guide_reporter_umi_counts_inferred_value.inferred_value, attribute_name))}

#
# Given the datastructure containing the inference results "observed_guide_reporter_umi_counts_inferred", iterate through the entire datastructure to generate
# another datastructure that contains the observed alleles (protospacer/surrogate/barcode) for each whitelist reporter in either a dictionary or dataframe format.
# This is the foundational datastructure used for analyzing the mutations for each whitelist reporter.
#
def get_matchset_alleleseries(observed_guide_reporter_umi_counts_inferred: GeneralMappingInferenceDict, attribute_name: str, contains_surrogate: bool, contains_barcode: bool, contains_umi: bool): 
    #
    #   DEFINE THE DEFAULTDICTS FOR COUNTING
    #
    ambiguous_ignored_umi_noncollapsed_alleledict : GeneralAlleleDict  = defaultdict(lambda: defaultdict(int))
    ambiguous_ignored_umi_collapsed_alleledict : GeneralAlleleDict  = defaultdict(lambda: defaultdict(int))
    ambiguous_ignored_alleledict : GeneralAlleleDict  = defaultdict(lambda: defaultdict(int))

    ambiguous_accepted_umi_noncollapsed_alleledict : GeneralAlleleDict = defaultdict(lambda: defaultdict(int))
    ambiguous_accepted_umi_collapsed_alleledict : GeneralAlleleDict = defaultdict(lambda: defaultdict(int))
    ambiguous_accepted_alleledict : GeneralAlleleDict = defaultdict(lambda: defaultdict(int))

    ambiguous_spread_umi_noncollapsed_alleledict : GeneralAlleleDict = defaultdict(lambda: defaultdict(int))
    ambiguous_spread_umi_collapsed_alleledict : GeneralAlleleDict = defaultdict(lambda: defaultdict(int))
    ambiguous_spread_alleledict : GeneralAlleleDict = defaultdict(lambda: defaultdict(int))

    #
    # ITERATE THROUGH THE NON-ERROR INFERRED RESULTS AND FILL THE COUNTS
    #
    inferred_value_results: InferenceResult
    for observed_sequence, inferred_value_results in get_non_error_dict(observed_guide_reporter_umi_counts_inferred, attribute_name).items():
        #
        #   Get the relevant attributes
        #
        observed_value_counts: Union[int, CounterType[Optional[str]]] = inferred_value_results.observed_value
        inferred_value_result: CompleteInferenceMatchResult =  inferred_value_results.inferred_value 
            
        match_set_single_inference_match_result : Optional[MatchSetSingleInferenceMatchResult] = getattr(inferred_value_result, attribute_name)
        assert match_set_single_inference_match_result is not None, "match_set_single_inference_match_result should not be none since this is from the non error list. Developer error."

        matches: pd.DataFrame = match_set_single_inference_match_result.value.matches
        if not matches.empty:
            # ITERATE THROUGH MATCHE(S) TO PERFORM COUNTS
            for whitelist_reporter_series in matches.iterrows(): 
                # UMI-BASED COUNTING
                whitelist_sequence_index = tuple(whitelist_reporter_series[1])
                observed_sequence_index = tuple(observed_sequence)
                if contains_umi:
                    assert isinstance(observed_value_counts, Counter), f"For UMI, expecting observed value is a Counter, but type is {type(observed_value_counts)}"
                    ambiguous_accepted_umi_noncollapsed_alleledict[whitelist_sequence_index][observed_sequence_index] += sum(observed_value_counts.values())
                    ambiguous_accepted_umi_collapsed_alleledict[whitelist_sequence_index][observed_sequence_index] += len(observed_value_counts.values())

                    ambiguous_spread_umi_noncollapsed_alleledict[whitelist_sequence_index][observed_sequence_index] += sum(observed_value_counts.values()) / float(matches.shape[0])
                    ambiguous_spread_umi_collapsed_alleledict[whitelist_sequence_index][observed_sequence_index] += len(observed_value_counts.values()) / float(matches.shape[0])
                    
                    # If there is no ambiguous matches, then add to ambiguous_ignored counter
                    if matches.shape[0] == 1:
                        ambiguous_ignored_umi_noncollapsed_alleledict[whitelist_sequence_index][observed_sequence_index] += sum(observed_value_counts.values())
                        ambiguous_ignored_umi_collapsed_alleledict[whitelist_sequence_index][observed_sequence_index] += len(observed_value_counts.values())

                # STANDARD NON-UMI BASED COUNTING
                else:
                    assert isinstance(observed_value_counts, int), f"For non UMI, expecting observed value is an int, but type is {type(observed_value_counts)}"
                    ambiguous_accepted_alleledict[whitelist_sequence_index][observed_sequence_index] += observed_value_counts
                    ambiguous_spread_alleledict[whitelist_sequence_index][observed_sequence_index] += observed_value_counts / float(matches.shape[0])

                    # If there is no ambiguous matches, then add to ambiguous_ignored counter
                    if matches.shape[0] == 1:
                        ambiguous_ignored_alleledict[whitelist_sequence_index][observed_sequence_index] += observed_value_counts

                        
    # Helper function that converts defaultdict to series
    def create_dict_counterseries(alleledict: GeneralAlleleDict) -> GeneralAlleleCountSeriesDict: 
        return {whitelist_sequence_key: pd.Series(observed_sequence_counterdict) for whitelist_sequence_key, observed_sequence_counterdict in alleledict.items()}
    
    def create_df_from_dict_counterseries(alleledict_counterseries: GeneralAlleleCountSeriesDict, contains_surrogate: bool, contains_barcode: bool):
        whitelist_reporter_columns = ["protospacer"] # For dynamically creating immutable tuple from reporter sequences based on if provided
                
        if contains_surrogate:
            whitelist_reporter_columns.append("surrogate")
        if contains_barcode:
            whitelist_reporter_columns.append("barcode")


        if alleledict_counterseries is not None:
            observed_reporter_df_counts_list = []
            for whitelist_reporter_tuple, observed_reporter_series_counts in alleledict_counterseries.items():
                observed_reporter_df_counts = pd.DataFrame(observed_reporter_series_counts)
                observed_reporter_df_counts.columns = ["counts"]
                observed_reporter_df_counts_index = observed_reporter_df_counts.index.to_frame()
                observed_reporter_df_counts_index.columns = ["observed_" + label for label in whitelist_reporter_columns]
                observed_reporter_df_counts = pd.concat([observed_reporter_df_counts, observed_reporter_df_counts_index], axis=1)
                observed_reporter_df_counts[["true_" + label for label in whitelist_reporter_columns]] = whitelist_reporter_tuple 
                observed_reporter_df_counts_list.append(observed_reporter_df_counts)
            observed_reporter_df_counts_total = pd.concat(observed_reporter_df_counts_list)
            return observed_reporter_df_counts_total
        else:
            return None
    #
    #   CONVERT THE COUNT DICTS INTO PANDAS SERIES, since this is a more ideal structure.
    #
    match_set_whitelist_reporter_observed_sequence_counter_series_results = MatchSetWhitelistReporterObservedSequenceCounterSeriesResults()
    match_set_whitelist_reporter_observed_sequence_counter_series_results.ambiguous_ignored_umi_noncollapsed_alleleseries_dict = create_dict_counterseries(ambiguous_ignored_umi_noncollapsed_alleledict)
    match_set_whitelist_reporter_observed_sequence_counter_series_results.ambiguous_ignored_umi_collapsed_alleleseries_dict = create_dict_counterseries(ambiguous_ignored_umi_collapsed_alleledict)
    match_set_whitelist_reporter_observed_sequence_counter_series_results.ambiguous_ignored_alleleseries_dict = create_dict_counterseries(ambiguous_ignored_alleledict)

    match_set_whitelist_reporter_observed_sequence_counter_series_results.ambiguous_accepted_umi_noncollapsed_alleleseries_dict = create_dict_counterseries(ambiguous_accepted_umi_noncollapsed_alleledict)
    match_set_whitelist_reporter_observed_sequence_counter_series_results.ambiguous_accepted_umi_collapsed_alleleseries_dict = create_dict_counterseries(ambiguous_accepted_umi_collapsed_alleledict)
    match_set_whitelist_reporter_observed_sequence_counter_series_results.ambiguous_accepted_alleleseries_dict = create_dict_counterseries(ambiguous_accepted_alleledict)

    match_set_whitelist_reporter_observed_sequence_counter_series_results.ambiguous_spread_umi_noncollapsed_alleleseries_dict = create_dict_counterseries(ambiguous_spread_umi_noncollapsed_alleledict)
    match_set_whitelist_reporter_observed_sequence_counter_series_results.ambiguous_spread_umi_collapsed_alleleseries_dict = create_dict_counterseries(ambiguous_spread_umi_collapsed_alleledict)
    match_set_whitelist_reporter_observed_sequence_counter_series_results.ambiguous_spread_alleleseries_dict = create_dict_counterseries(ambiguous_spread_alleledict)

    #
    #    Create the dataframes based on the countdicts
    #
    match_set_whitelist_reporter_observed_sequence_counter_series_results.ambiguous_ignored_umi_noncollapsed_allele_df = create_df_from_dict_counterseries(match_set_whitelist_reporter_observed_sequence_counter_series_results.ambiguous_ignored_umi_noncollapsed_alleleseries_dict, contains_surrogate=contains_surrogate, contains_barcode=contains_barcode)
    match_set_whitelist_reporter_observed_sequence_counter_series_results.ambiguous_ignored_umi_collapsed_allele_df = create_df_from_dict_counterseries(match_set_whitelist_reporter_observed_sequence_counter_series_results.ambiguous_ignored_umi_collapsed_alleleseries_dict, contains_surrogate=contains_surrogate, contains_barcode=contains_barcode)
    match_set_whitelist_reporter_observed_sequence_counter_series_results.ambiguous_ignored_allele_df = create_df_from_dict_counterseries(match_set_whitelist_reporter_observed_sequence_counter_series_results.ambiguous_ignored_alleleseries_dict, contains_surrogate=contains_surrogate, contains_barcode=contains_barcode)

    match_set_whitelist_reporter_observed_sequence_counter_series_results.ambiguous_accepted_umi_noncollapsed_allele_df = create_df_from_dict_counterseries(match_set_whitelist_reporter_observed_sequence_counter_series_results.ambiguous_accepted_umi_noncollapsed_alleleseries_dict, contains_surrogate=contains_surrogate, contains_barcode=contains_barcode)
    match_set_whitelist_reporter_observed_sequence_counter_series_results.ambiguous_accepted_umi_collapsed_allele_df = create_df_from_dict_counterseries(match_set_whitelist_reporter_observed_sequence_counter_series_results.ambiguous_accepted_umi_collapsed_alleleseries_dict, contains_surrogate=contains_surrogate, contains_barcode=contains_barcode)
    match_set_whitelist_reporter_observed_sequence_counter_series_results.ambiguous_accepted_allele_df = create_df_from_dict_counterseries(match_set_whitelist_reporter_observed_sequence_counter_series_results.ambiguous_accepted_alleleseries_dict, contains_surrogate=contains_surrogate, contains_barcode=contains_barcode)

    match_set_whitelist_reporter_observed_sequence_counter_series_results.ambiguous_spread_umi_noncollapsed_allele_df = create_df_from_dict_counterseries(match_set_whitelist_reporter_observed_sequence_counter_series_results.ambiguous_spread_umi_noncollapsed_alleleseries_dict, contains_surrogate=contains_surrogate, contains_barcode=contains_barcode)
    match_set_whitelist_reporter_observed_sequence_counter_series_results.ambiguous_spread_umi_collapsed_allele_df = create_df_from_dict_counterseries(match_set_whitelist_reporter_observed_sequence_counter_series_results.ambiguous_spread_umi_collapsed_alleleseries_dict, contains_surrogate=contains_surrogate, contains_barcode=contains_barcode)
    match_set_whitelist_reporter_observed_sequence_counter_series_results.ambiguous_spread_allele_df = create_df_from_dict_counterseries(match_set_whitelist_reporter_observed_sequence_counter_series_results.ambiguous_spread_alleleseries_dict, contains_surrogate=contains_surrogate, contains_barcode=contains_barcode)
        
    return match_set_whitelist_reporter_observed_sequence_counter_series_results














from copy import deepcopy
import numpy as np
# TODO: Move to package

def pad_series(series, possible_max = 0):
    max_len = max(series.apply(len).max(), possible_max)
    return series.apply(lambda item: item.ljust(max_len, 'X')), max_len

def pad_sequence(sequence, sequence_max=0):
    return sequence.ljust(sequence_max, "X")

# Copying function from crispr-millipede package - don't want import entire package for a single function
def get_substitution_encoding(true_sequence, observed_sequence, skip_index=0):
    assert len(true_sequence) == len(observed_sequence), f"Sequences not same length. True={true_sequence}, Observed={observed_sequence}" # Ensure the aligned sequence (from allele table) is equal size to the reference sequence
    
    nucleotides = ["A","C","T","G","N","-", "X"] # List of possible nucleotides
    encodings_per_position = []
    mismatch_mappings_per_position = []
    for index in range(0, len(true_sequence)): # Iterate through each base and check for substitution
        # TODO Ensure sequences are uppercase
        nucleotides_mm = nucleotides[:]
        nucleotides_mm.remove(true_sequence[index])
        mm_encoding = pd.Series(np.repeat(0, len(nucleotides)-1))
        if observed_sequence[index] == true_sequence[index]: # If the observed sequence is same as reference
            pass
        else:
            mm_index = nucleotides_mm.index(observed_sequence[index])
            mm_encoding[mm_index] = 1
        mismatch_mappings_per_position.append(nucleotides_mm)
        encodings_per_position.append(mm_encoding)

    encodings_per_position_df = pd.DataFrame(encodings_per_position).T
    mismatch_mappings_per_position_df = pd.DataFrame(mismatch_mappings_per_position).T

    encodings_per_position_df.columns = list(true_sequence)
    mismatch_mappings_per_position_df.columns = list(true_sequence)

    mismatch_mappings_per_position_POS_list = np.arange(mismatch_mappings_per_position_df.shape[1]).repeat(mismatch_mappings_per_position_df.shape[0])
    mismatch_mappings_per_position_REF_list = np.asarray(list(true_sequence)).repeat(mismatch_mappings_per_position_df.shape[0]).astype(np.object_)
    mismatch_mappings_per_position_ALT_list = mismatch_mappings_per_position_df.T.values.flatten()
    mismatch_mappings_per_position_full_list = mismatch_mappings_per_position_POS_list.astype(np.str_).astype(object)+mismatch_mappings_per_position_REF_list + np.repeat(">", len(mismatch_mappings_per_position_REF_list)) + mismatch_mappings_per_position_ALT_list
    encodings_per_position_list = encodings_per_position_df.T.values.flatten()

    
    # Encodings per position DF, mismatch mappings per position DF, encodings per position flattened, mismatch mappings per position flattened, mismatch mapping position in flattened list, mismatch mapping ref in flattened list, mismatch mapping alt in flattened list, all substitutions made
    
    index = pd.MultiIndex.from_tuples(zip(mismatch_mappings_per_position_full_list, mismatch_mappings_per_position_POS_list, mismatch_mappings_per_position_REF_list, mismatch_mappings_per_position_ALT_list), names=["FullChange", "Position","Ref", "Alt"])
    
    assert len(encodings_per_position_list) == len(index), f"{len(encodings_per_position_list)} == {len(index)}"
    encodings_per_position_series = pd.Series(encodings_per_position_list, index = index, name="encoding")
    return encodings_per_position_series


def determine_mutations_in_sequence(true_sequence, observed_sequence):
    mutation_array = []
    lowest_len = min(len(true_sequence), len(observed_sequence))
    for i in range(lowest_len):
        if true_sequence[i] != observed_sequence[i]:
            preceding_nt_context = None
            succeeding_nt_context = None
            ref_nt = true_sequence[i]
            alt_nt = observed_sequence[i]
            if i == 0:
                preceding_nt_context = "_"
                succeeding_nt_context = true_sequence[i+1]
            elif i == len(true_sequence)-1:
                preceding_nt_context = true_sequence[i-1]
                succeeding_nt_context = "_"
            else:
                preceding_nt_context = true_sequence[i-1]
                succeeding_nt_context = true_sequence[i+1]
                
            mutation_series = pd.Series((preceding_nt_context, ref_nt, alt_nt,succeeding_nt_context, i), index=["preceding_nt_context", "ref_nt", "alt_nt", "succeeding_nt_context", "position"])
            mutation_array.append(mutation_series)
    observed_sequence_mutation_df = pd.DataFrame(mutation_array)
    return observed_sequence_mutation_df


def get_mutation_profile(match_set_whitelist_reporter_observed_sequence_counter_series_results: MatchSetWhitelistReporterObservedSequenceCounterSeriesResults, whitelist_reporter_df: pd.DataFrame, contains_surrogate: bool, contains_barcode: bool) -> MatchSetWhitelistReporterObservedSequenceMutationProfiles: 
    
    # Function to generate unlinked mutations for particular count type
    def generate_mutations_results(alleleseries: Optional[GeneralAlleleCountSeriesDict], whitelist_reporter_df: pd.DataFrame, contains_surrogate: bool, contains_barcode: bool) -> Optional[ObservedSequenceMutationProfile]:
        if alleleseries is not None:
            
            linked_mutations_whitelist_reporter_dict = {}
            all_observed_protospacer_unlinked_mutations_df_list = []
            if contains_surrogate:
                all_observed_surrogate_unlinked_mutations_df_list = []
            if contains_barcode:
                all_observed_barcode_unlinked_mutations_df_list = []

            for whitelist_reporter_sequence in whitelist_reporter_df.iterrows():
                #
                # GET OBSERVED SEQUENCE FRROM WHITELIST TUPLE KEY (the index of whitelist_reporter_df may have the tuple, though it may not be provided. I also want to ensure that the order of protospacer/surrogate/barcode is maintained if columns are swapped)
                #
                whitelist_sequence_pretuple_list = [] # For dynamically creating immutable tuple from reporter sequences based on if provided
                whitelist_protospacer_sequence = whitelist_reporter_sequence[1]["protospacer"][:]
                whitelist_sequence_pretuple_list.append(whitelist_protospacer_sequence)
                if contains_surrogate:
                    whitelist_surrogate_sequence = whitelist_reporter_sequence[1]["surrogate"][:]
                    whitelist_sequence_pretuple_list.append(whitelist_surrogate_sequence)
                if contains_barcode:
                    whitelist_barcode_sequence = whitelist_reporter_sequence[1]["barcode"][:]
                    whitelist_sequence_pretuple_list.append(whitelist_barcode_sequence)
                whitelist_reporter_tuple = tuple(whitelist_sequence_pretuple_list)
                
                linked_mutations_series_list = []
                try:
                    
                    # CALL THE DICT WITH WHITELIST TUPLE
                    observed_sequences_df = alleleseries[whitelist_reporter_tuple] # Get the observed sequences (DF) for the specific whitelist sequence
                    
                    
                    # PAD THE WHITELIST AND OBSERVED SEQUENCES TO ALLOW ENCODING OF EQUAL LENGTH SEQUENCES
                    whitelist_reporter_sequence_copy = deepcopy(whitelist_reporter_sequence) # Copy the sequences since it will be modified by padding
                    padded_observed_sequences_results: List[List[str], Optional[List[str]], Optional[List[str]]] = []
                    for level in range(observed_sequences_df.index.nlevels): # Iterate over sequence components (protospacer or surrogate or barcode)
                        series_padding_results = pad_series(observed_sequences_df.index.get_level_values(level).to_series(), possible_max=len(whitelist_reporter_sequence_copy[1].iloc[level])) # Pad the observed sequence series (protospacer or surrogate or barcode)
                        padded_observed_sequences_results.append(series_padding_results[0].to_list()) # Append the results to the result list
                        whitelist_reporter_sequence_copy[1].iloc[level] = pad_sequence(whitelist_reporter_sequence_copy[1].iloc[level], sequence_max=series_padding_results[1]) # Pad the result
                    padded_observed_sequence_results_tuples = [tuple(padded_observed_sequence_list) for padded_observed_sequence_list in zip(*padded_observed_sequences_results)]
                    observed_sequences_df.index = pd.MultiIndex.from_tuples(padded_observed_sequence_results_tuples)
                    
                    
                    for observed_sequences, count in observed_sequences_df.items(): # Iterate through each observed sequence to tally mutations
                        # Get protospacer unlinked mutations
                        observed_protospacer_sequence = observed_sequences[0]
                        observed_protospacer_unlinked_mutations_df = determine_mutations_in_sequence(true_sequence=whitelist_reporter_sequence_copy[1]["protospacer"], observed_sequence=observed_protospacer_sequence) # Get DF of mutations in 
                        observed_protospacer_unlinked_mutations_df["count"] = count # Add count
                        observed_protospacer_unlinked_mutations_df.loc[:, whitelist_reporter_sequence_copy[1].index.values] = whitelist_reporter_sequence_copy[1].values # Annotate the DF with the whitelist sequences
                        all_observed_protospacer_unlinked_mutations_df_list.append(observed_protospacer_unlinked_mutations_df) # Add to complete list
                        
                        # Get protospacer linked mutations 
                        observed_linked_mutations_series = get_substitution_encoding(true_sequence=whitelist_reporter_sequence_copy[1]["protospacer"], observed_sequence=observed_protospacer_sequence) # Get DF of mutations in 
                        observed_linked_mutations_series = pd.concat({'protospacer': observed_linked_mutations_series}, names=['SequenceType'])

                        # Get surrogate mutations (if surrogate provided)
                        if contains_surrogate:
                            observed_surrogate_sequence = observed_sequences[1]
                            observed_surrogate_unlinked_mutations_df = determine_mutations_in_sequence(true_sequence=whitelist_reporter_sequence_copy[1]["surrogate"], observed_sequence=observed_surrogate_sequence) # Get DF of mutations in 
                            observed_surrogate_unlinked_mutations_df["count"] = count # Add count
                            observed_surrogate_unlinked_mutations_df.loc[:, whitelist_reporter_sequence_copy[1].index.values] = whitelist_reporter_sequence_copy[1].values# Annotate the DF with the whitelist sequences
                            all_observed_surrogate_unlinked_mutations_df_list.append(observed_surrogate_unlinked_mutations_df)# Add to complete list
                            
                            observed_surrogate_linked_mutations_series = get_substitution_encoding(true_sequence=whitelist_reporter_sequence_copy[1]["surrogate"], observed_sequence=observed_surrogate_sequence) # Get DF of mutations in 
                            observed_surrogate_linked_mutations_series = pd.concat({'surrogate': observed_surrogate_linked_mutations_series}, names=['SequenceType'])
                            observed_linked_mutations_series = pd.concat([observed_linked_mutations_series,observed_surrogate_linked_mutations_series])
                            
                        
                        # Get barcode mutations (if barcode provided)
                        if contains_barcode:
                            observed_barcode_sequence = observed_sequences[2]
                            observed_barcode_unlinked_mutations_df = determine_mutations_in_sequence(true_sequence=whitelist_reporter_sequence_copy[1]["barcode"], observed_sequence=observed_barcode_sequence) # Get DF of mutations in 
                            observed_barcode_unlinked_mutations_df["count"] = count # Add count
                            observed_barcode_unlinked_mutations_df.loc[:, whitelist_reporter_sequence_copy[1].index.values] = whitelist_reporter_sequence_copy[1].values # Annotate the DF with the whitelist sequences
                            all_observed_barcode_unlinked_mutations_df_list.append(observed_barcode_unlinked_mutations_df) # Add to complete list
                            
                            observed_barcode_linked_mutations_series = get_substitution_encoding(true_sequence=whitelist_reporter_sequence_copy[1]["barcode"], observed_sequence=observed_barcode_sequence) # Get DF of mutations in 
                            observed_barcode_linked_mutations_series = pd.concat({'barcode': observed_barcode_linked_mutations_series}, names=['SequenceType'])
                            observed_linked_mutations_series = pd.concat([observed_linked_mutations_series,observed_barcode_linked_mutations_series])
                        
                        observed_linked_mutations_series["count"] = count # Add count
                        linked_mutations_series_list.append(observed_linked_mutations_series)
                    
                    # Add linked mutations DF to dict
                    linked_mutations_df = pd.concat(linked_mutations_series_list, axis=1).transpose()
                    linked_mutations_df.index = observed_sequences_df.index # Set the index of the encodings to be the observed sequences
                    linked_mutations_whitelist_reporter_dict.update({whitelist_reporter_tuple:linked_mutations_df})
                except KeyError as e:
                    print(f"No observed sequences found for: {whitelist_reporter_tuple}")
                
                
                
            all_observed_protospacer_unlinked_mutations_df = pd.concat(all_observed_protospacer_unlinked_mutations_df_list) # Concat list of all whitelist mutations to sequence dataframe
            observed_sequence_mutations = ObservedSequenceMutationProfile(all_observed_protospacer_unlinked_mutations_df=all_observed_protospacer_unlinked_mutations_df, linked_mutations_whitelist_reporter_dict=linked_mutations_whitelist_reporter_dict) # Initialize results object with groupby results
            
            # Add surrogate result to result object (if provided)
            if contains_surrogate:
                all_observed_surrogate_unlinked_mutations_df = pd.concat(all_observed_surrogate_unlinked_mutations_df_list)
                observed_sequence_mutations.all_observed_surrogate_unlinked_mutations_df=all_observed_surrogate_unlinked_mutations_df
                
            # Add barcode result to result object (if provided)
            if contains_barcode:
                all_observed_barcode_unlinked_mutations_df = pd.concat(all_observed_barcode_unlinked_mutations_df_list)
                observed_sequence_mutations.all_observed_barcode_unlinked_mutations_df=all_observed_barcode_unlinked_mutations_df
            
            return observed_sequence_mutations
        else:
            return None
    
    # Generate mutation counts for each count type
    mutations_results = MatchSetWhitelistReporterObservedSequenceMutationProfiles()
    print("Generating ambiguous_ignored_umi_noncollapsed_mutations")
    mutations_results.ambiguous_ignored_umi_noncollapsed_mutations = generate_mutations_results(match_set_whitelist_reporter_observed_sequence_counter_series_results.ambiguous_ignored_umi_noncollapsed_alleleseries_dict, whitelist_reporter_df, contains_surrogate, contains_barcode)
    print("Generating ambiguous_ignored_umi_collapsed_mutations")
    mutations_results.ambiguous_ignored_umi_collapsed_mutations = generate_mutations_results(match_set_whitelist_reporter_observed_sequence_counter_series_results.ambiguous_ignored_umi_collapsed_alleleseries_dict, whitelist_reporter_df, contains_surrogate, contains_barcode)
    print("Generating ambiguous_ignored_unlinked_mutations")
    mutations_results.ambiguous_ignored_mutations = generate_mutations_results(match_set_whitelist_reporter_observed_sequence_counter_series_results.ambiguous_ignored_alleleseries_dict, whitelist_reporter_df, contains_surrogate, contains_barcode)

    print("Generating ambiguous_accepted_umi_noncollapsed_mutations")
    mutations_results.ambiguous_accepted_umi_noncollapsed_mutations = generate_mutations_results(match_set_whitelist_reporter_observed_sequence_counter_series_results.ambiguous_accepted_umi_noncollapsed_alleleseries_dict, whitelist_reporter_df, contains_surrogate, contains_barcode)
    print("Generating ambiguous_accepted_umi_collapsed_mutations")
    mutations_results.ambiguous_accepted_umi_collapsed_mutations = generate_mutations_results(match_set_whitelist_reporter_observed_sequence_counter_series_results.ambiguous_accepted_umi_collapsed_alleleseries_dict, whitelist_reporter_df, contains_surrogate, contains_barcode)
    print("Generating ambiguous_accepted_mutations")
    mutations_results.ambiguous_accepted_mutations = generate_mutations_results(match_set_whitelist_reporter_observed_sequence_counter_series_results.ambiguous_accepted_alleleseries_dict, whitelist_reporter_df, contains_surrogate, contains_barcode)

    print("Generating ambiguous_spread_umi_noncollapsed_mutations")
    mutations_results.ambiguous_spread_umi_noncollapsed_mutations = generate_mutations_results(match_set_whitelist_reporter_observed_sequence_counter_series_results.ambiguous_spread_umi_noncollapsed_alleleseries_dict, whitelist_reporter_df, contains_surrogate, contains_barcode)
    print("Generating ambiguous_spread_umi_collapsed_mutations")
    mutations_results.ambiguous_spread_umi_collapsed_mutations = generate_mutations_results(match_set_whitelist_reporter_observed_sequence_counter_series_results.ambiguous_spread_umi_collapsed_alleleseries_dict, whitelist_reporter_df, contains_surrogate, contains_barcode)
    print("Generating ambiguous_spread_mutations")
    mutations_results.ambiguous_spread_mutations = generate_mutations_results(match_set_whitelist_reporter_observed_sequence_counter_series_results.ambiguous_spread_alleleseries_dict, whitelist_reporter_df, contains_surrogate, contains_barcode)
    
    return mutations_results

 
def tally_linked_mutation_count_per_sequence(mutations_results: MatchSetWhitelistReporterObservedSequenceMutationProfiles, contains_surrogate: bool, contains_barcode: bool, count_attribute_name: str="ambiguous_accepted_umi_noncollapsed_mutations") -> LinkedMutationCounters:
    protospacer_total_mutation_counter: CounterType = Counter()

    surrogate_total_mutation_counter: Optional[CounterType] = None
    barcode_total_mutation_counter: Optional[CounterType] = None
    if contains_surrogate:
        surrogate_total_mutation_counter = Counter()
    if contains_barcode:
        barcode_total_mutation_counter = Counter()

    # TODO: THIS IS HARDCODED TO ambiguous_accepted_umi_noncollapsed
        
    mutation_results_typed = getattr(mutations_results, count_attribute_name)
    if mutation_results_typed is not None:
        for linked_mutations_whitelist_reporter_df in mutation_results_typed.linked_mutations_whitelist_reporter_dict.values():
            accetable_SNP_columns = (linked_mutations_whitelist_reporter_df.columns.get_level_values("SequenceType") != "count") & (linked_mutations_whitelist_reporter_df.columns.get_level_values("Ref") != "X") & (linked_mutations_whitelist_reporter_df.columns.get_level_values("Alt") != "X")


            protospacer_columns = (linked_mutations_whitelist_reporter_df.columns.get_level_values("SequenceType") == "protospacer")
            protospacer_mutation_count_per_allele_series = linked_mutations_whitelist_reporter_df.loc[:, protospacer_columns & accetable_SNP_columns].sum(axis=1)
            for allele_index, mutation_count in enumerate(protospacer_mutation_count_per_allele_series):
                protospacer_total_mutation_counter[mutation_count] += linked_mutations_whitelist_reporter_df["count"][allele_index]

            if contains_surrogate:
                surrogate_columns = (linked_mutations_whitelist_reporter_df.columns.get_level_values("SequenceType") == "surrogate")
                surrogate_mutation_count_per_allele_series = linked_mutations_whitelist_reporter_df.loc[:, surrogate_columns & accetable_SNP_columns].sum(axis=1)
                for allele_index, mutation_count in enumerate(surrogate_mutation_count_per_allele_series):
                    surrogate_total_mutation_counter[mutation_count] += linked_mutations_whitelist_reporter_df["count"][allele_index]

            if contains_barcode:
                barcode_columns = (linked_mutations_whitelist_reporter_df.columns.get_level_values("SequenceType") == "barcode")
                barcode_mutation_count_per_allele_series = linked_mutations_whitelist_reporter_df.loc[:, barcode_columns & accetable_SNP_columns].sum(axis=1)
                for allele_index, mutation_count in enumerate(barcode_mutation_count_per_allele_series):
                    barcode_total_mutation_counter[mutation_count] += linked_mutations_whitelist_reporter_df["count"][allele_index]
    
    return LinkedMutationCounters(protospacer_total_mutation_counter=protospacer_total_mutation_counter,
                                surrogate_total_mutation_counter=surrogate_total_mutation_counter,
                                barcode_total_mutation_counter=barcode_total_mutation_counter)



from matplotlib import pyplot as plt
import numpy as np
import seaborn as sns
import pandas as pd
import anndata as ad
from ..models.editing_models import MatchSetWhitelistReporterObservedSequenceMutationProfiles
from Bio.Seq import Seq
from typing import List, Tuple, Optional, Dict
import matplotlib.patches as mpatches
from matplotlib.font_manager import FontProperties
from matplotlib.patches import Rectangle


def show_reporter_encoding_frequency(encoding_df, sequence_type, count_column_name="count"):
    encoding_df_count = encoding_df.loc[:, count_column_name]
    encoding_df_type = encoding_df.loc[:, encoding_df.columns.get_level_values("SequenceType") == sequence_type]
    encoding_frequency_series = encoding_df_type.mul(encoding_df_count, axis=0).sum(axis=0) / encoding_df_count.sum() * 100
    return encoding_frequency_series

def plot_editing_frequency_heatmap(replicate_scores, protospacer_start_coordinate:int, protospacer_size:int = 20, title: str=""):
    # Group by Position and sum the frequencies
    frequency_dfs = []
    for replicate_score in replicate_scores:
        frequency_df = pd.DataFrame(replicate_score, columns=["Frequency"]).groupby(['Position', 'Ref'])["Frequency"].sum().reset_index()
        frequency_dfs.append(frequency_df)

    # Create a single-row DataFrame for the heatmap
    heatmap_data = pd.DataFrame(columns=range(frequency_dfs[0].shape[0]))
    for rep_i, frequency_df in enumerate(frequency_dfs):
        heatmap_data.loc[rep_i] = [0.0] * frequency_df.shape[0]  # Initialize with zeros

    # Add the editing frequencies to the heatmap input dataframe
    for index, _ in frequency_dfs[0].iterrows():
        position = frequency_dfs[0].loc[index]['Position']
        for rep_i, frequency_df in enumerate(frequency_dfs):
            heatmap_data.at[rep_i, position] = frequency_df.loc[index]['Frequency']

    # Prepare annotations (only show greater than 5% editing)
    new_annot = heatmap_data.copy()
    new_annot[new_annot <= 5] = ""

    # Custom function to format numbers to ".2f" and leave strings unchanged
    def format_numbers(val):
        if isinstance(val, (int, float)):
            return f'{val:.1f}%'
        return val

    # Apply the custom function to each element in the DataFrame
    new_annot = new_annot.applymap(format_numbers)

    # Plot the heatmap using seaborn
    plt.figure(figsize=(35, 1.5))
    heatmap = sns.heatmap(heatmap_data, cmap='Reds', annot=new_annot, fmt='', cbar=True, vmin=0, vmax=100,
                          linecolor='grey', linewidths=1, cbar_kws={"shrink": 0.7, "pad": 0.01} )
    plt.xlabel('Position')
    plt.ylabel('Replicate')
    plt.title('Editing Frequencies Heatmap')

    # Set reference bases as labels
    bold_font = FontProperties(weight='bold')
    heatmap.set_xticklabels(frequency_df["Ref"], fontproperties=bold_font)
    heatmap.tick_params(axis='x', labelsize=16)

    heatmap.set_yticklabels(heatmap.get_yticklabels(), rotation=0)

    rect = Rectangle((protospacer_start_coordinate, 0.00), protospacer_size, 1, fill=False, edgecolor='black', linewidth=3)

    heatmap.add_patch(rect)
    sns.despine()
    plt.title(title)
    plt.show()

def plot_lfc_scatter_plot(anndata: ad.AnnData, 
                          reference_sequence_dict: Dict[str, Seq],
                          groupby_columns_list: List[str], 
                          enriched_population_label: str, 
                          baseline_population_label: str, 
                          population_column_name:str, 
                          count_size_population_label: Optional[str] = None,
                          guide_type_column_name:str = "type",
                          guide_targeting_type_label: str = "targeting",
                          is_igRNA_column_name: Optional[str] = "is_igRNA", 
                          strand_column_name: Optional[str] = "strand", 
                          coordinate_column_name: str = "coordinate", # Set to hg38_editsite_coordinate
                          chromosome_column_name: str = "chromosome",
                          score_layer_name: str ="lfc",
                          count_size_max_threshold: int = 300,
                          figure_width: int = 100,
                          figure_height: int = 6,
                          count_to_scale_factor: int = 0.5
                          ):
    for group in anndata.obs.groupby(groupby_columns_list):
        enriched_pop_sample_df = group[1][group[1][population_column_name] == enriched_population_label]
        assert enriched_pop_sample_df.shape[0] == 1, f"Multiple samples matching enriched condition in group {group[1]}"
        enriched_pop_sample = enriched_pop_sample_df
        
        baseline_pop_sample_df = group[1][group[1][population_column_name] == baseline_population_label]
        assert baseline_pop_sample_df.shape[0] == 1, f"Multiple samples matching enriched condition in group {group[1]}"
        baseline_pop_sample = baseline_pop_sample_df

        if count_size_population_label is not None:
            count_size_pop_sample_df = group[1][group[1][population_column_name] == count_size_population_label]
            assert count_size_pop_sample_df.shape[0] == 1, f"Multiple samples matching condition to calculate count size in group {group[1]}"
            count_size_pop_sample = count_size_pop_sample_df
        
        enriched_anndata = anndata[enriched_pop_sample.index, :] # DEVELOPER NOTE: Will thow error of indices of samples are changed from ints
        baseline_anndata = anndata[baseline_pop_sample.index, :] # DEVELOPER NOTE: Will thow error of indices of samples are changed from ints
        if count_size_population_label is not None:
            count_size_anndata = anndata[count_size_pop_sample.index, :] # DEVELOPER NOTE: Will thow error of indices of samples are changed from ints
        
        
        enriched_targeting_anndata = enriched_anndata[:, enriched_anndata.var[guide_type_column_name] == guide_targeting_type_label]
        baseline_targeting_anndata = baseline_anndata[:, baseline_anndata.var[guide_type_column_name] == guide_targeting_type_label]
        if count_size_population_label is not None:
            count_size_targeting_anndata = count_size_anndata[:, count_size_anndata.var[guide_type_column_name] == guide_targeting_type_label]

        # Get igRNAs and pgRNtargeting_anndataAs separately
        is_igRNA = None
        is_pgRNA = None
        if is_igRNA_column_name is not None:
            is_igRNA = enriched_targeting_anndata.var[is_igRNA_column_name]
            is_pgRNA = ~is_igRNA
        else:
            # If igRNA not provided, set all guides as pgRNA by default
            is_igRNA = pd.Series([False] * len(enriched_targeting_anndata), index=enriched_targeting_anndata.index)
            is_pgRNA = pd.Series([True] * len(enriched_targeting_anndata), index=enriched_targeting_anndata.index)

        positive_strand = None
        negative_strand = None
        if strand_column_name is not None:
            positive_strand = enriched_targeting_anndata.var[strand_column_name] == "+"
            negative_strand = ~positive_strand
        else:
            # If strand not provided, set all guides as positive strand by default
            negative_strand = pd.Series([False] * len(enriched_targeting_anndata), index=enriched_targeting_anndata.index)
            positive_strand = pd.Series([True] * len(enriched_targeting_anndata), index=enriched_targeting_anndata.index)

        # Get the guide coordinates
        coordinates = enriched_targeting_anndata.var[coordinate_column_name]

        chromosome = enriched_targeting_anndata.var[chromosome_column_name][0]
        ids_pgRNA = enriched_targeting_anndata[:, is_pgRNA].var.index
        positive_strand_pgRNA = positive_strand[is_pgRNA] 

        coordinates_pgRNA = coordinates[is_pgRNA.values]
        scores_pgRNA = enriched_targeting_anndata[:, is_pgRNA].layers[score_layer_name].mean(axis=0).flatten()
        if count_size_population_label is not None:
            counts_pgRNA = count_size_targeting_anndata[:, is_pgRNA].X[0]
        else:
            counts_pgRNA = enriched_targeting_anndata[:, is_pgRNA].X[0] + baseline_targeting_anndata[:, is_pgRNA].X[0]
        counts_pgRNA[counts_pgRNA>count_size_max_threshold] = count_size_max_threshold

        coordinates_igRNA = coordinates[is_igRNA]
        ids_igRNA = enriched_targeting_anndata[:, is_igRNA].var.index
        scores_igRNA = enriched_targeting_anndata[:, is_igRNA].layers[score_layer_name].mean(axis=0).flatten()
        if count_size_population_label is not None:
            counts_pgRNA = count_size_targeting_anndata[:, is_igRNA].X[0]
        else:
            counts_igRNA = enriched_targeting_anndata[:, is_igRNA].X[0] + baseline_targeting_anndata[:, is_igRNA].X[0]
        counts_igRNA[counts_igRNA>count_size_max_threshold] = count_size_max_threshold
        
        positive_strand_igRNA = positive_strand[is_igRNA] 

        fig, axes = plt.subplots(1, 1, figsize=(figure_width, figure_height), sharex=True)
        edge_colors = ['black', 'orange']
        
        scatter = axes.scatter(coordinates_pgRNA, scores_pgRNA, alpha=0.7, c="blue", edgecolors=np.where(positive_strand_pgRNA, edge_colors[0], edge_colors[1]), label="pgRNA", s=counts_pgRNA*count_to_scale_factor)
        if is_igRNA_column_name is not None:
            axes.scatter(coordinates_igRNA, scores_igRNA, alpha=0.3, c="black", edgecolors=np.where(positive_strand_pgRNA, edge_colors[0], edge_colors[1]), label="igRNA", s=counts_igRNA*count_to_scale_factor)

        ylim = axes.get_ylim()
        axes.set_ylim(ylim[0], ylim[1]+1)
        xlim = axes.get_xlim()
        for coordinate in range(int(xlim[0])+2,int(xlim[1])-1):
            axes.text(coordinate-1, ylim[0]+0.1, reference_sequence_dict[chromosome][coordinate], fontsize=10, alpha=1)

        # Create legends for point sizes, edge colors, and dot colors
        if is_igRNA_column_name is not None:
            dotcolor_legend = axes.legend(title="Type", loc="upper right")
            plt.gca().add_artist(dotcolor_legend)
        if strand_column_name:
            edgecolor_patches = [mpatches.Patch(color=color, label=["+", "-"][i]) for i, color in enumerate(edge_colors)]
            edgecolor_legend = plt.legend(handles=edgecolor_patches, title="Strand (Edge Color)", loc="upper right", bbox_to_anchor=(1.0, 0.83))
            plt.gca().add_artist(edgecolor_legend)

        if count_size_population_label is not None:
            size_title = f"{count_size_population_label} Count"
        else:
            size_title = "Total Count"
        size_legend = plt.legend(*scatter.legend_elements(prop="sizes", num=5, func=lambda x: x * 1/count_to_scale_factor), title=size_title, loc="upper right", bbox_to_anchor=(1.0, 0.65))
        plt.gca().add_artist(size_legend)
        
        fig.suptitle("_".join([str(term) for term in group[0]]))
        plt.show() 


def plot_mutation_count_histogram(mutation_counter, title="Bar Plot of Counts Sorted by Index", filename: Optional[str] = None):
    fig, ax = plt.subplots(1, figsize=(5, 5))

    if sum(mutation_counter) > 0: # If there exists mutation counts
        sorted_mutation_counter = sorted(mutation_counter.items())

        # Extract keys (indexes) and values from the sorted Counter
        indexes, counts = zip(*sorted_mutation_counter)

        # Create a bar plot
        
        ax.bar(indexes, counts, color='blue')

        # Add labels and title
        ax.set_xlabel('Mutations Per Allele')
        ax.set_ylabel('Read Count')
    
    fig.suptitle(title)
    if filename is None:
        plt.show()
    else:
        fig.savefig(filename)



def plot_trinucleotide_mutational_signature(mutations_results: MatchSetWhitelistReporterObservedSequenceMutationProfiles, label: str, count_attribute_name="ambiguous_accepted_umi_noncollapsed_mutations", unlinked_mutation_attribute_name = "all_observed_surrogate_unlinked_mutations_df", possible_refs: List[str] = ["A","C","G","T"], possible_alts: List[str] = ["A","C","G","T","X","N","-","X"], empty_context_character: str = "_", filename: Optional[str] = None):
    fig, ax = plt.subplots(1, figsize=(15, 5*1))

    mutation_results_typed = getattr(mutations_results, count_attribute_name)
    if mutation_results_typed is not None:
        unlinked_mutations_df: pd.DataFrame = getattr(mutation_results_typed, unlinked_mutation_attribute_name)

        if (unlinked_mutations_df is not None) and (unlinked_mutations_df.shape[0] > 0): # If there are mutations
            
            '''
                Pre-processing mutation data
            '''
            # Prepare unique trinucleotide contexts based on provided REF and ALT sequences
            unique_trinucleotide_contexts_series_list: List[pd.Series] = []
            for ref_nt in possible_refs:
                for alt_nt in possible_alts:
                    if ref_nt != alt_nt:
                        for preceding_ref_nt in possible_refs + [empty_context_character]:
                            for succeeding_ref_nt in possible_refs + [empty_context_character]:
                                unique_trinucleotide_contexts_series_list.append(pd.Series([preceding_ref_nt, ref_nt, alt_nt, succeeding_ref_nt], index=["preceding_nt_context", "ref_nt", "alt_nt", "succeeding_nt_context"]))
            unique_trinucleotide_contexts_df = pd.DataFrame(unique_trinucleotide_contexts_series_list)

            # Calculate the counts for each trinucleotide context
            unique_trinucleotide_contexts_counts_series_list: List[pd.Series] = []
            for mononucleotide_context_indices, mononucleotide_context_group in unique_trinucleotide_contexts_df.groupby(["ref_nt", "alt_nt"]):
                for row_index, row in mononucleotide_context_group.iterrows():
                    row["count"] = unlinked_mutations_df.loc[(unlinked_mutations_df["preceding_nt_context"] == row["preceding_nt_context"]) & (unlinked_mutations_df["ref_nt"] == row["ref_nt"]) & (unlinked_mutations_df["alt_nt"] == row["alt_nt"]) & (unlinked_mutations_df["succeeding_nt_context"] == row["succeeding_nt_context"]), "count"].sum()
                    unique_trinucleotide_contexts_counts_series_list.append(row)
            unique_trinucleotide_contexts_counts_df = pd.DataFrame(unique_trinucleotide_contexts_counts_series_list)

            '''
                Prepare constants for plotting
            '''
            # Define constants normalization and axis setting
            unlinked_mutations_df_groupby_trinucleotide_context_summed = unlinked_mutations_df.groupby(["preceding_nt_context", "ref_nt", "alt_nt", "succeeding_nt_context"]).sum("count")
            total_mutations = unlinked_mutations_df_groupby_trinucleotide_context_summed["count"].sum()
            max_context_frequency = (np.round(max(unlinked_mutations_df_groupby_trinucleotide_context_summed["count"]/total_mutations) * 100)) + 1 # Get the max context frequency to set plot y-axis limit
            
            '''
                Generation annotations of each group and their positions.
            '''
            current_position = 0
            variant_class_positions: List[Tuple[str, int]] = []
            
            # TODO: Do I need to sort by preceeding and succeeding context to ensure order is correct? Check later by printing
            
            # Collect the variant position labels
            unique_trinucleotide_contexts_counts_df_groupby_mononucleotide_context = unique_trinucleotide_contexts_counts_df.groupby(["ref_nt", "alt_nt"])
            for mononucleotide_context_indices, mononucleotide_context_group in unique_trinucleotide_contexts_counts_df_groupby_mononucleotide_context:
                position = current_position + (mononucleotide_context_group.shape[0]/2)
                current_position = current_position + mononucleotide_context_group.shape[0]
                variant_class_positions.append((f"{mononucleotide_context_indices[0]}>{mononucleotide_context_indices[1]}", position)) # Add the label 
            
            '''
                Perform plotting
            '''
            color_palette = sns.color_palette("Set2",unique_trinucleotide_contexts_counts_df_groupby_mononucleotide_context.ngroups)
            bar_colors = np.asarray([color_palette[group_index] for group_index, (_, mononucleotide_context_group) in enumerate(unique_trinucleotide_contexts_counts_df_groupby_mononucleotide_context) for _ in range(mononucleotide_context_group.shape[0])])
            variant_frequencies = pd.concat([(mononucleotide_context_group["count"]/total_mutations)*100 for mononucleotide_context_indices, mononucleotide_context_group in unique_trinucleotide_contexts_counts_df_groupby_mononucleotide_context])
            variant_frequencies.plot.bar(x='lab', y='val', rot=0, color = bar_colors, ax=ax)
            ax.set_ylim(0, max_context_frequency)
            ax.set_xticks([position for variant_class, position in variant_class_positions], [variant_class for variant_class, _ in variant_class_positions])
            ax.set_ylabel("Proportion of mutations (%)")
            ax.set_xlabel("Variant Type")
            ax.set_title(label)
    
    if filename is None:
        plt.show()    
    else:
        fig.savefig(filename)
    
def plot_positional_mutational_signature(mutations_results: MatchSetWhitelistReporterObservedSequenceMutationProfiles, label: str, count_attribute_name="ambiguous_accepted_umi_noncollapsed_mutations", unlinked_mutation_attribute_name = "all_observed_surrogate_unlinked_mutations_df", possible_refs: List[str] = ["A","C","G","T"], possible_alts: List[str] = ["A","C","G","T","X","N","-","X"], empty_context_character: str = "_", min_position: int = 0, max_position: Optional[int] = None, filename: Optional[str] = None):
    fig, ax = plt.subplots(1, figsize=(15, 5*1))

    mutation_results_typed = getattr(mutations_results, count_attribute_name)
    if mutation_results_typed is not None:
        unlinked_mutations_df: pd.DataFrame = getattr(mutation_results_typed, unlinked_mutation_attribute_name)

        if (unlinked_mutations_df is not None) and (unlinked_mutations_df.shape[0] > 0): # If there are mutations
            '''
                Pre-processing mutation data
            '''
            max_position = int(unlinked_mutations_df.position.quantile(q=0.95)) if max_position is None else max_position
            unique_positional_contexts_series_list: List[pd.Series] = []
            for ref_nt in possible_refs:
                for alt_nt in possible_alts:
                    if ref_nt != alt_nt:
                        for position in range(0, max_position):
                            unique_positional_contexts_series_list.append(pd.Series([ref_nt, alt_nt, position], index=["ref_nt", "alt_nt", "position"]))
            unique_positional_contexts_df = pd.DataFrame(unique_positional_contexts_series_list)

            unique_positional_contexts_counts_series_list: List[pd.Series] = []
            for mononucleotide_context_indices, mononucleotide_context_group in unique_positional_contexts_df.groupby(["ref_nt", "alt_nt"]):
                for row_index, row in mononucleotide_context_group.iterrows():
                    row["count"] = unlinked_mutations_df.loc[(unlinked_mutations_df["ref_nt"] == row["ref_nt"]) & (unlinked_mutations_df["alt_nt"] == row["alt_nt"]) & (unlinked_mutations_df["position"] == row["position"]), "count"].sum()
                    unique_positional_contexts_counts_series_list.append(row)
            unique_positional_contexts_counts_df = pd.DataFrame(unique_positional_contexts_counts_series_list)

            '''
                Prepare constants for plotting
            '''
            # Define constants normalization and axis setting
            unlinked_mutations_df_groupby_positional_context_summed = unlinked_mutations_df.groupby(["ref_nt", "alt_nt", "position"]).sum("count")
            total_mutations = unlinked_mutations_df_groupby_positional_context_summed["count"].sum()
            max_context_frequency = (np.round(max(unlinked_mutations_df_groupby_positional_context_summed["count"]/total_mutations) * 100)) + 1 # Get the max context frequency to set plot y-axis limit
            
            '''
                Generation annotations of each group and their positions.
            '''
            current_position = 0
            variant_class_positions: List[Tuple[str, int]] = []
            
            # TODO: Do I need to sort by preceeding and succeeding context to ensure order is correct? Check later by printing
            
            # Collect the variant position labels
            unique_positional_contexts_counts_df_groupby_mononucleotide_context = unique_positional_contexts_counts_df.groupby(["ref_nt", "alt_nt"])
            for mononucleotide_context_indices, mononucleotide_context_group in unique_positional_contexts_counts_df_groupby_mononucleotide_context:
                position = current_position + (mononucleotide_context_group.shape[0]/2)
                current_position = current_position + mononucleotide_context_group.shape[0]
                variant_class_positions.append((f"{mononucleotide_context_indices[0]}>{mononucleotide_context_indices[1]}", position)) # Add the label 
            
            '''
                Perform plotting
            '''
            color_palette = sns.color_palette("Set2",unique_positional_contexts_counts_df_groupby_mononucleotide_context.ngroups)
            bar_colors = np.asarray([color_palette[group_index] for group_index, (_, mononucleotide_context_group) in enumerate(unique_positional_contexts_counts_df_groupby_mononucleotide_context) for _ in range(mononucleotide_context_group.shape[0])])
            variant_frequencies = pd.concat([(mononucleotide_context_group["count"]/total_mutations)*100 for mononucleotide_context_indices, mononucleotide_context_group in unique_positional_contexts_counts_df_groupby_mononucleotide_context])
            variant_frequencies.plot.bar(x='lab', y='val', rot=0, color = bar_colors, ax=ax)
            ax.set_ylim(0, max_context_frequency)
            ax.set_xticks([position for variant_class, position in variant_class_positions], [variant_class for variant_class, _ in variant_class_positions])
            ax.set_ylabel("Proportion of mutations (%)")
            ax.set_xlabel("Variant Type")
            ax.set_title(label)
    
    if filename is None:
        plt.show()    
    else:
        fig.savefig(filename)


#!/usr/bin/env python

from . import mapping 
from . import utility
from . import models
from . import processing
from . import visualization
from . import quality_control
from . import postprocessing


###
### GUIDE PARSING HELPER FUNCTIONS
###
from typeguard import typechecked
from typing import Union, Optional, List, Tuple, DefaultDict
from typing import Counter as CounterType
from ..models.mapping_models import ProtospacerCounter, ProtospacerDictUMICounter, ProtospacerBarcodeCounter, ProtospacerBarcodeDictUMICounter, ProtospacerSurrogateCounter, ProtospacerSurrogateDictUMICounter, ProtospacerSurrogateBarcodeCounter, ProtospacerSurrogateBarcodeDictUMICounter
from collections import Counter
from typing import Callable
from functools import partial
from datetime import datetime
import gzip
import re
from Bio.Seq import Seq
from collections import Counter, defaultdict

# This is for grouping the FASTQ lines in 1 tuple.
def grouper(iterable, n=4):
    "s -> (s0,s1,...sn-1), (sn,sn+1,...s2n-1), (s2n,s2n+1,...s3n-1), ..."
    return zip(*[iter(iterable)]*n)

@typechecked
def get_standard_observed_sequence_counts(  fastq_r1_fn: str, 
                                            fastq_r2_fn: Optional[str], 
                                            
                                            protospacer_pattern_regex: Optional[str],
                                            surrogate_pattern_regex: Optional[str],
                                            barcode_pattern_regex: Optional[str],
                                            umi_pattern_regex: Optional[str],

                                            protospacer_left_flank:Optional[str],
                                            protospacer_right_flank:Optional[str],
                                            protospacer_start_position:Optional[int],
                                            protospacer_end_position:Optional[int],
                                            protospacer_length: Optional[int],

                                            surrogate_left_flank:Optional[str],
                                            surrogate_right_flank:Optional[str],
                                            surrogate_start_position:Optional[int],
                                            surrogate_end_position:Optional[int],
                                            surrogate_length: Optional[int],

                                            barcode_left_flank:Optional[str],
                                            barcode_right_flank:Optional[str],
                                            barcode_start_position:Optional[int],
                                            barcode_end_position:Optional[int],
                                            barcode_length: Optional[int],

                                            umi_left_flank:Optional[str],
                                            umi_right_flank:Optional[str],
                                            umi_start_position:Optional[int],
                                            umi_end_position:Optional[int],
                                            umi_length: Optional[int],

                                            is_protospacer_r1: Optional[bool], 
                                            is_surrogate_r1: Optional[bool], 
                                            is_barcode_r1: Optional[bool],
                                            is_umi_r1: Optional[bool],
                                            is_protospacer_header: Optional[bool], 
                                            is_surrogate_header: Optional[bool], 
                                            is_barcode_header: Optional[bool],
                                            is_umi_header: Optional[bool],
                                        
                                            revcomp_protospacer: Optional[bool], 
                                            revcomp_surrogate: Optional[bool], 
                                            revcomp_barcode: Optional[bool],
                                            revcomp_umi: Optional[bool],
                                            contains_surrogate: bool,
                                            contains_barcode: bool,
                                            contains_umi: bool) -> Union[CounterType[str], 
                                                                        DefaultDict[str, CounterType[str]], 
                                                                        CounterType[Tuple[str, str]], 
                                                                        DefaultDict[Tuple[str, str], CounterType[str]], 
                                                                        CounterType[Tuple[str, str, str]], 
                                                                        DefaultDict[Tuple[str, str, str], CounterType[str]]]:
    revcomp = lambda sequence, do_revcomp: str(Seq(sequence).reverse_complement()) if do_revcomp else sequence

    def parse_sequence(
            template_sequence: str,
            sequence_pattern_regex: Optional[str],
            sequence_left_flank:Optional[str],
            sequence_right_flank:Optional[str],
            sequence_start_position:Optional[int],
            sequence_end_position:Optional[int],
            sequence_length: Optional[int],
            revcomp_sequence: Optional[bool],
            sequence_type: str):
        
        if sequence_pattern_regex is not None:
            """
                Parse entirely by regex
            """
            sequence = re.search(sequence_pattern_regex, template_sequence).group(1).rstrip()  
        else:
            """
                Get sequence position start
            """
            if sequence_left_flank is not None:
                """
                    Parse by left flank
                """
                position_start = template_sequence.find(sequence_left_flank) + len(sequence_left_flank)
            elif sequence_start_position is not None:
                """
                    Use provided position
                """
                position_start = sequence_start_position
            else:
                raise Exception(f"{sequence_type}_left_flank or {sequence_type}_start_position must be provided")
            



            """
                Get sequence position end
            """
            if sequence_right_flank is not None:
                """
                    Parse by left flank
                """
                position_end = template_sequence.find(sequence_right_flank)
            elif sequence_end_position is not None:
                """
                    Use provided position
                """
                position_end = sequence_end_position
            elif sequence_length is not None:
                """
                    Use provided sequence length
                """
                position_end = position_start + sequence_length
            else:
                raise Exception(f"{sequence_type}_right_flank or {sequence_type}_end_position or {sequence_type}_length must be provided")
            
            sequence = template_sequence[position_start:position_end].rstrip() # rstrip is important, since need to remove newlines if sequence_length exceeds template_sequence length
        
        # Return sequence (reverse complement if necessary)
        assert revcomp_sequence is not None, f"revcomp_{sequence_type} must be provided, does the sequence need to be reverse complemented?"
        return revcomp(sequence, revcomp_sequence)

    protospacer_parse_sequence_partial = partial(parse_sequence, 
                                                 sequence_pattern_regex=protospacer_pattern_regex,
                                                 sequence_left_flank=protospacer_left_flank,
                                                 sequence_right_flank=protospacer_right_flank,
                                                 sequence_start_position=protospacer_start_position,
                                                 sequence_end_position=protospacer_end_position,
                                                 sequence_length=protospacer_length,
                                                 revcomp_sequence=revcomp_protospacer,
                                                 sequence_type="protospacer")
    surrogate_parse_sequence_partial = partial(parse_sequence, 
                                                 sequence_pattern_regex=surrogate_pattern_regex,
                                                 sequence_left_flank=surrogate_left_flank,
                                                 sequence_right_flank=surrogate_right_flank,
                                                 sequence_start_position=surrogate_start_position,
                                                 sequence_end_position=surrogate_end_position,
                                                 sequence_length=surrogate_length,
                                                 revcomp_sequence=revcomp_surrogate,
                                                 sequence_type="surrogate")
    barcode_parse_sequence_partial = partial(parse_sequence, 
                                                 sequence_pattern_regex=barcode_pattern_regex,
                                                 sequence_left_flank=barcode_left_flank,
                                                 sequence_right_flank=barcode_right_flank,
                                                 sequence_start_position=barcode_start_position,
                                                 sequence_end_position=barcode_end_position,
                                                 sequence_length=barcode_length,
                                                 revcomp_sequence=revcomp_barcode,
                                                 sequence_type="barcode")
    umi_parse_sequence_partial = partial(parse_sequence, 
                                                 sequence_pattern_regex=umi_pattern_regex,
                                                 sequence_left_flank=umi_left_flank,
                                                 sequence_right_flank=umi_right_flank,
                                                 sequence_start_position=umi_start_position,
                                                 sequence_end_position=umi_end_position,
                                                 sequence_length=umi_length,
                                                 revcomp_sequence=revcomp_umi,
                                                 sequence_type="umi")
    
    def get_sequence_from_read_group(fastq_read_group, parse_sequence_partial, is_r1:bool, is_header: bool, includes_r2: bool, sequence_type: str):
                if includes_r2:
                    r1_header_read, _ = fastq_read_group[0]
                    r1_sequence_read, r2_sequence_read = fastq_read_group[1]
                else:
                    r1_header_read = fastq_read_group[0]
                    r1_sequence_read = fastq_read_group[1]


                if is_r1 is True:
                    sequence = parse_sequence_partial(template_sequence=r1_sequence_read)
                else:
                    if is_header:
                        sequence = parse_sequence_partial(template_sequence=r1_header_read)
                    else:
                        if includes_r2:
                            sequence = parse_sequence_partial(template_sequence=r2_sequence_read)
                        else:
                            raise Exception(f"{sequence_type} cannot be parsed from R1 since is_{sequence_type}_r1 is False, is_{sequence_type}_header is False, and R2 is not provided")

                return sequence
    get_protospacer_from_read_group_partial = partial(get_sequence_from_read_group,
                                                      parse_sequence_partial=protospacer_parse_sequence_partial, 
                                                      is_r1=is_protospacer_r1, 
                                                      is_header=is_protospacer_header,
                                                      sequence_type="protospacer")
    get_surrogate_from_read_group_partial = partial(get_sequence_from_read_group,
                                                      parse_sequence_partial=surrogate_parse_sequence_partial, 
                                                      is_r1=is_surrogate_r1, 
                                                      is_header=is_surrogate_header,
                                                      sequence_type="surrogate")
    get_barcode_from_read_group_partial = partial(get_sequence_from_read_group,
                                                      parse_sequence_partial=barcode_parse_sequence_partial, 
                                                      is_r1=is_barcode_r1, 
                                                      is_header=is_barcode_header,
                                                      sequence_type="barcode")
    get_umi_from_read_group_partial = partial(get_sequence_from_read_group,
                                                      parse_sequence_partial=umi_parse_sequence_partial, 
                                                      is_r1=is_umi_r1, 
                                                      is_header=is_umi_header,
                                                      sequence_type="umi")
    
    def parse_fastq(fastq_r1_filehandler, fastq_r2_filehandler):
        if fastq_r2_fn is None: # ONLY R1
            fastq_single_read_grouper = grouper(fastq_r1_filehandler)
            if not contains_barcode: # ONLY R1; NO BARCODE
                if not contains_umi: # ONLY R1; NO BARCODE; NO UMI
                    if not contains_surrogate:
                        sequence_counter: ProtospacerCounter = Counter()
                        for fastq_single_read_group in fastq_single_read_grouper:
                            protospacer = get_protospacer_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                            sequence_counter[protospacer] += 1
                    else:
                        sequence_counter: ProtospacerSurrogateCounter = Counter()
                        for fastq_single_read_group in fastq_single_read_grouper:
                            protospacer = get_protospacer_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                            surrogate = get_surrogate_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                            sequence_counter[(protospacer, surrogate)] += 1

                else: # ONLY R1; NO BARCODE; YES UMI
                    if not contains_surrogate:
                        sequence_counter: ProtospacerDictUMICounter = defaultdict(Counter)
                        for fastq_single_read_group in fastq_single_read_grouper:
                            protospacer = get_protospacer_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                            umi = get_umi_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                            sequence_counter[protospacer][umi] += 1
                    else:
                        sequence_counter: ProtospacerSurrogateDictUMICounter = defaultdict(Counter)
                        for fastq_single_read_group in fastq_single_read_grouper:
                            protospacer = get_protospacer_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                            surrogate = get_surrogate_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                            umi = get_umi_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                            sequence_counter[(protospacer, surrogate)][umi] += 1
            else: # ONLY R1; YES BARCODE
                if not contains_umi: # ONLY R1; YES BARCODE; NO UMI
                    if not contains_surrogate:
                        sequence_counter: ProtospacerBarcodeCounter = Counter()
                        for fastq_single_read_group in fastq_single_read_grouper:
                            protospacer = get_protospacer_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                            barcode = get_barcode_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                            sequence_counter[(protospacer, barcode)] += 1
                    else:
                        sequence_counter: ProtospacerSurrogateBarcodeCounter = Counter()
                        for fastq_single_read_group in fastq_single_read_grouper:
                            protospacer = get_protospacer_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                            surrogate = get_surrogate_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                            barcode = get_barcode_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                            sequence_counter[(protospacer, surrogate, barcode)] += 1
                else: # ONLY R1; YES BARCODE; YES UMI
                    if not contains_surrogate:
                        sequence_counter: ProtospacerBarcodeDictUMICounter = defaultdict(Counter)
                        for fastq_single_read_group in fastq_single_read_grouper:
                            protospacer = get_protospacer_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                            barcode = get_barcode_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                            umi = get_umi_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                            sequence_counter[(barcode, protospacer)][umi] += 1
                    else:
                        sequence_counter: ProtospacerSurrogateBarcodeDictUMICounter = defaultdict(Counter)
                        for fastq_single_read_group in fastq_single_read_grouper:
                            protospacer = get_protospacer_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                            surrogate = get_surrogate_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                            barcode = get_barcode_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                            umi = get_umi_from_read_group_partial(fastq_read_group=fastq_single_read_group, includes_r2=False)
                            sequence_counter[(barcode, surrogate, protospacer)][umi] += 1
        else: # YES R2
            fastq_paired_read_grouper = grouper(zip(fastq_r1_filehandler, fastq_r2_filehandler))
            if not contains_barcode: # YES R2; NO BARCODE
                if not contains_umi: # YES R2; NO BARCODE; NO UMI
                    if not contains_surrogate:
                        sequence_counter: ProtospacerCounter = Counter()
                        for fastq_paired_read_group in fastq_paired_read_grouper:
                            protospacer = get_protospacer_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                            if contains_surrogate:
                                sequence_counter[protospacer] += 1
                    else:
                        sequence_counter: ProtospacerSurrogateCounter = Counter()
                        for fastq_paired_read_group in fastq_paired_read_grouper:
                            protospacer = get_protospacer_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                            surrogate = get_surrogate_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                            sequence_counter[(protospacer, surrogate)] += 1
                else: # YES R2; NO BARCODE; YES UMI
                    if not contains_surrogate:
                        sequence_counter: ProtospacerDictUMICounter = defaultdict(Counter)
                        for fastq_paired_read_group in fastq_paired_read_grouper:
                            protospacer = get_protospacer_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                            umi = get_umi_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                            sequence_counter[protospacer][umi] += 1
                    else:
                        sequence_counter: ProtospacerSurrogateDictUMICounter = defaultdict(Counter)
                        for fastq_paired_read_group in fastq_paired_read_grouper:
                            protospacer = get_protospacer_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                            surrogate = get_surrogate_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                            umi = get_umi_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                            sequence_counter[(protospacer, surrogate)][umi] += 1
            
            else: # YES R2; YES BARCODE
                if umi_pattern_regex is None: # YES R2; YES BARCODE; NO UMI
                    if not contains_surrogate:
                        sequence_counter: ProtospacerSurrogateBarcodeCounter = Counter()
                        for fastq_paired_read_group in fastq_paired_read_grouper:
                            protospacer = get_protospacer_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                            barcode = get_barcode_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                            sequence_counter[(protospacer, barcode)] += 1
                    else:
                        sequence_counter: ProtospacerSurrogateBarcodeCounter = Counter()
                        for fastq_paired_read_group in fastq_paired_read_grouper:
                            protospacer = get_protospacer_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                            surrogate = get_surrogate_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                            barcode = get_barcode_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                            sequence_counter[(protospacer, surrogate, barcode)] += 1
                else: # YES R2; YES BARCODE; YES UMI
                    if not contains_surrogate:
                        sequence_counter: ProtospacerSurrogateBarcodeDictUMICounter = defaultdict(Counter)
                        for fastq_paired_read_group in fastq_paired_read_grouper:
                            protospacer = get_protospacer_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                            barcode = get_barcode_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                            umi = get_umi_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                            sequence_counter[(protospacer, barcode)][umi] += 1
                    else:
                        sequence_counter: ProtospacerSurrogateBarcodeDictUMICounter = defaultdict(Counter)
                        for fastq_paired_read_group in fastq_paired_read_grouper:
                            protospacer = get_protospacer_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                            surrogate = get_surrogate_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                            barcode = get_barcode_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                            umi = get_umi_from_read_group_partial(fastq_read_group=fastq_paired_read_group, includes_r2=True)
                            sequence_counter[(protospacer, surrogate, barcode)][umi] += 1
        return sequence_counter
    
    before_file_loading_time = datetime.now()
    # Open R1 file handler
    fastq_r1_filehandler = None
    if fastq_r1_fn.endswith('.gz'):
        print(f"Opening FASTQ.gz file with gzip, filename={fastq_r1_fn}")
        fastq_r1_filehandler = gzip.open(fastq_r1_fn, "rt", encoding="utf-8")
    else:
        print(f"Opening FASTQ file, filename={fastq_r1_fn}")
        fastq_r1_filehandler = open(fastq_r1_fn, "r")
    
    # Open R2 file handler if provided
    fastq_r2_filehandler = None
    if fastq_r2_fn is not None:
        if fastq_r2_fn.endswith('.gz'):
            print(f"Opening FASTQ.gz file with gzip, filename={fastq_r2_fn}")
            fastq_r2_filehandler = gzip.open(fastq_r2_fn, "rt", encoding="utf-8")
        else:
            print(f"Opening FASTQ file, filename={fastq_r2_fn}")
            fastq_r2_filehandler = open(fastq_r2_fn, "r")
    after_file_loading_time = datetime.now()
    print(f"{(after_file_loading_time-before_file_loading_time).seconds} seconds for file loading")

    sequence_counter = parse_fastq(fastq_r1_filehandler, fastq_r2_filehandler)
    after_parsing_time = datetime.now()
    print(f"{(after_parsing_time-after_file_loading_time).seconds} seconds for parsing")

    # Close the file handlers when done
    if fastq_r1_filehandler is not None:
        fastq_r1_filehandler.close()
    if fastq_r2_filehandler is not None:
        fastq_r2_filehandler.close()

    return sequence_counter


from typing import Counter as CounterType

def calculate_average_editing_frequency(total_mutation_counter: CounterType) -> float:
    total_editing_counts = (sum(total_mutation_counter.values()) - total_mutation_counter[0])
    total_counts = sum(total_mutation_counter.values())
    if total_counts > 0:
        editing_efficiency = total_editing_counts/total_counts
        return editing_efficiency
    else:
        return 0


###
### GUIDE MAPPING HELPER FUNCTIONS
###
import numpy as np

# Vector encoding of each base, since we want to vectorized
non_ambigious_encoding_dict = dict({
    "A": np.asarray([1,0,0,0]),
    "C": np.asarray([0,1,0,0]),
    "G": np.asarray([0,0,1,0]),
    "T": np.asarray([0,0,0,1]),
    "U": np.asarray([0,0,0,1]),
    "X": np.asarray([9999,9999,9999,9999]) # NOTE: This represents a padding token for sequences. Set to some high number, so that the calculated hamming distance is very high and won't be selected (which will happen if the observed sequence is greater than the whitelist sequence)
})

'''
 Helper function for retrieving the encoding for ambigious bases based on IUPAC codes
'''
def ambiguity_encoder(bases):
    return np.logical_or.reduce([non_ambigious_encoding_dict[base] for base in bases]).astype(int)

'''
    Final dictionary for getting encoding of each IUPAC base
'''
full_encoding_dict = dict({
    "A": non_ambigious_encoding_dict["A"],
    "C": non_ambigious_encoding_dict["C"],
    "G": non_ambigious_encoding_dict["G"],
    "T": non_ambigious_encoding_dict["T"], 
    "X": non_ambigious_encoding_dict["X"], # NOTE: This represents a padding token for sequences 
    "R": ambiguity_encoder(["A", "G"]),
    "Y": ambiguity_encoder(["C", "T"]),
    "S": ambiguity_encoder(["G", "C"]),
    "W": ambiguity_encoder(["A", "T"]),
    "K": ambiguity_encoder(["G", "T"]),
    "M": ambiguity_encoder(["A", "C"]),
    "B": ambiguity_encoder(["C", "G", "T"]),
    "D": ambiguity_encoder(["A", "G", "T"]),
    "H": ambiguity_encoder(["A", "C", "T"]),
    "V": ambiguity_encoder(["A", "C", "G"]),
    "N": ambiguity_encoder(["A", "C", "G", "T"]),
})

'''
    Main function to encode a single base
'''
def encode_DNA_base(char):
    return full_encoding_dict[char]
encode_DNA_base_vectorized = np.vectorize(encode_DNA_base, signature='()->(n)') # Vectorized function for a string (i.e. gRNA)

'''
    Function for converting string (i.e. gRNA) into a np array of chars  - may be deprecated (NOTE 20221202)
'''
def numpify_string(string):
    return np.array(list(string), dtype=str)
numpify_string_vectorized = np.vectorize(numpify_string, signature='()->(n)') # Vectorize the function

def encode_guide_series(guide_series) -> np.array:
    guide_numpy = guide_series.to_numpy(dtype=object)
    guide_numpy = guide_numpy.astype(str)
    guide_numpy_char = np.array(list(map(list, guide_numpy))) # Map into a list of list of characters
    guide_numpy_encoding = encode_DNA_base_vectorized(guide_numpy_char)
    return guide_numpy_encoding

def retrieve_hamming_distance_whitelist(target_guide_encoded, whitelist_guide_encoded):
    '''
        This takes a encoded guide sequence and a list of encoded whitelisted guides and matrix computes the hamming distance of the 
        encoded guide across all whitelisted guides in a single operation
        
        (target_guide_encoded*whitelist_guide_encoded[:, np.newaxis]).sum(axis=3) # Determines 
        
    '''
    return ((target_guide_encoded*whitelist_guide_encoded[:, np.newaxis]).sum(axis=3)^1).sum(axis=2).flatten()

# Deprecated
def determine_hamming_distance_classic(seq1, seq2):
    if len(seq1) != len(seq2):
        raise ValueError(f"Sequences must have equal length. {seq1} and {seq2}")
    
    distance = 0
    for i in range(len(seq1)):
        if seq1[i] != seq2[i]:
            distance += 1
    
    return distance
